/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "canvas.h"
#include "canvasoperation.h"
#include "iOpenGLHelper.h"
#include <QMutexLocker>

static const char * g_err_MAXLAYERS   = "Maximum layers count.";
static const char * g_err_WRONGIDX    = "Index is out of range.";
static const char * g_err_BUFTOOSMALL = "Buffer is too small.";
static const char * g_err_NOCONTEXT   = "OpenGL context is not available.";


ICanvas * g_fac_SPAWN_CANVAS (QWidget * parent, int width, int height, IOpenGLHelper * helper) {
    return new Canvas (parent, width, height, helper);
}

ICanvas * g_fac_SPAWN_CANVAS (int width, int height, IOpenGLHelper * helper) {
    return new Canvas (NULL, width, height, helper);
}

float CanvasOperation::bleedingRate (void) const {
    return MAX (0.0f,
                // Fast strokes have lesser bleeding rate.
                m_Source->bleedingRate () *
                (1.0f - 0.05f * sqrtf (m_Velocity[0] * m_Velocity[0] +
                                       m_Velocity[1] * m_Velocity[1]))
            );
}

float CanvasOperation::suckingRate (void) const {
    return MIN (1.0f,
                // Viscous paints' sucking rate is greater.
                (1.0f + 0.25f * m_PaintingStyle->paintViscosity ()) *
                 m_Source->suckingRate ()
           ) * m_FeedbackMu;
}

float CanvasOperation::stipplingSensitivity (void) const {
    return MATH::CLAMP (m_Instrument->textureSensitivityModifier () *
                        m_Source->stipplingSensitivity () +
                        0.20f * m_PaintingStyle->paintViscosity (),
                        0.01f,
                        1.00f);
}


/// Constructor. =)
Canvas::Canvas (QWidget * parent, int width, int height, IOpenGLHelper * helper) :
        QGLWidget (QGLFormat (
                   QGL::NoSampleBuffers |
                   QGL::NoStereoBuffers |
                   QGL::DoubleBuffer |
                // QGL::SingleBuffer |
                   QGL::Rgba |
                   QGL::NoDepthBuffer |
                   QGL::NoAccumBuffer |
                   QGL::NoStencilBuffer
                             ),
                   parent)
{
    m_Logger         = NULL;
    m_RenderingChain = NULL;
    m_CurrentLayer   = NULL;
    m_CurrentChannel = NULL;
    m_Instrument     = NULL;
    m_Connector      = NULL;
    m_LastError      = NULL;
    m_PaintingStyle  = NULL;
    m_MixerLayer     = NULL;
    m_PipelineDry    = NULL;
    m_ActionControl  = NULL;
    m_LoggingEnabled = FALSE;
    m_HasGLContext   = FALSE;
    m_Changed        = FALSE;
    m_MixerVisible   = FALSE;
    m_StrokeStarted  = FALSE;
    m_TabletPresence = FALSE;
    m_PlaneWidth     = width;
    m_PlaneHeight    = height;
    m_Helper         = helper;
    m_MixerWidth     = 256;
    m_MixerHeight    = 256;
    m_MixerOpacity   = 1.0f;
    m_ImageDpi       = 100;
    if (m_Helper) {
        m_Helper->addRef ();
    }

    m_SmartStroke_mu = 1.0f;
    m_UseSmartStroke = TRUE;

    m_SplineFeedbackMu = NULL;
    m_LastReservoirSz  = -1.0f;

    m_CurrentStrokeLength_x = 0;
    m_CurrentStrokeLength_y = 0;

    //m_MixerVisible   = TRUE;

    m_LastDevUid            = 0;
    m_LastPointer           = 0;
    m_LineStrokeLength[0]   = 0;
    m_LineStrokeLength[1]   = 0;
    m_GlobalStatus          = QHASGlobal::STATUS_IDLE;
    m_LastPosInside_w       = FALSE;
    m_CumulativeRectTrigger = FALSE;
    m_CurrentTransaction    = NULL;

    m_DragMode       = DRAGSTATE_NOOP;
    m_LastXPos       = -10;
    m_LastYPos       = -10;
    m_LastXPos_w     = -10;
    m_LastYPos_w     = -10;

    m_Id = ISpawnedObject::allocateId ();
    ISpawnedObject::_register (this);

    // "Downward" gravity vector by default.
    m_Gravity[0] = 0.0f;
    m_Gravity[1] = 0.0f;

    // Identity transform by default.
    m_ViewTranslate[0] = 0.0f;
    m_ViewTranslate[1] = 0.0f;
    m_ViewTranslate[2] = 0.0f;
    m_ViewRotate[0]    = 0.0f;
    m_ViewRotate[1]    = 0.0f;
    m_ViewRotate[2]    = 0.0f;
    m_ViewScale[0]     = 1.0f;
    m_ViewScale[1]     = 1.0f;
    m_ViewScale[2]     = 1.0f;

    m_propMetaDataScale = 0.5f;

    m_StrokeDirection[0] = 0.0f;
    m_StrokeDirection[1] = 0.0f;
    m_PressureWRng[0]    = 0.0f;
    m_PressureWRng[1]    = 1.0f;

    // Create a papermap object.
    // FIXME Adjustable paper dpi?
    m_Papermap = new Papermap (helper, 100);
    rebuildMatrix ();

    m_rViewport.left     = 0;
    m_rViewport.top      = 0;
    m_rViewport.right    = this->width ()  - 1;
    m_rViewport.bottom   = this->height () - 1;
    m_rTargetClip.left   = 0;
    m_rTargetClip.top    = 0;
    m_rTargetClip.right  = width  - 1;
    m_rTargetClip.bottom = height - 1;

    // Init the widget-specific properties.
    setAutoFillBackground (FALSE);
    setMouseTracking      (TRUE);
    setMinimumSize        (128, 128);

    m_rPointer = QRect (0, 0, 0, 0);

    // Set the default mouse actions.
    m_MouseHash.insert (Qt::LeftButton,  DRAGSTATE_PAINT);
    m_MouseHash.insert (Qt::RightButton, DRAGSTATE_CANVAS_MOVE);

    // Reset the request map.
    for (int i = 0; i < RQ_CEILING; i++)
        m_ReqMap[i] = FALSE;

    QObject::connect (&m_timerConsciousness, SIGNAL(timeout()), this, SLOT(poke()));
    m_timerConsciousness.start (1024);
    m_timerUpdate.restart ();
}

Canvas::~Canvas (void) {
    deInit ();
    ISpawnedObject::_unregister (this);

  //logMessage (LOGIMPORTANCE_NEVERMIND, "Notifying everyone that the context is lost.");
  //emit glcontextLost();

    SAFE_RELEASE (m_Helper);

    logMessage (LOGIMPORTANCE_NEVERMIND, "Back to Sansara...");
    SAFE_RELEASE (m_Logger);
}

void Canvas::deInit (void) {
    if (m_RenderingChain) {
        m_RenderingChain->clear ();
    }

    foreach (ICanvasLayer * layer, m_Layers) {
        layer->release ();
    }

    m_Layers.clear ();
    SAFE_DELETE  (m_Papermap);
    SAFE_RELEASE (m_SplineFeedbackMu);
    SAFE_RELEASE (m_MixerLayer);
    SAFE_RELEASE (m_RenderingChain);
    SAFE_RELEASE (m_Instrument);
    SAFE_RELEASE (m_PipelineDry);
    SAFE_RELEASE (m_Connector);
}

void Canvas::resize (long width, long height) {
    m_PlaneWidth  = width;
    m_PlaneHeight = height;

    // Resize all layers.
    foreach (ICanvasLayer * layer, m_Layers) {
        layer->resize (width, height);
    }

    rebound ();
    updateRenderingChain ();
}

void Canvas::connect (IConnector * caller) {
    m_LastError = NULL;
    SAFE_RELEASE (m_Connector);

    if (caller) {
        m_Connector = caller;
        m_Connector->addRef ();

        QObject::connect (dynamic_cast<QObject*>(m_Connector), SIGNAL(disconnected()), this, SLOT(disconnected()));

        // Request the pipelines list.
        m_Connector->allocateObject (this, RQ_PIPELINES, ENTITY_PIPELINE_PLUGIN);
    }
}

void Canvas::setUniformScale (float scale) {
    m_ViewScale[0] = MATH::CLAMP (scale, 0.25f, 8.0f);
    m_ViewScale[1] = MATH::CLAMP (scale, 0.25f, 8.0f);
    m_ViewScale[2] = MATH::CLAMP (scale, 0.25f, 8.0f);
    rebuildMatrix ();
    rebound ();
    update ();
}

float Canvas::getUniformScale (void) const {
    return 0.5f * (m_ViewScale[0] + m_ViewScale[1]);
}

void Canvas::multiplyScale (float scale) {
    m_ViewScale[0] = MATH::CLAMP (m_ViewScale[0] * scale, 0.25f, 8.0f);
    m_ViewScale[1] = MATH::CLAMP (m_ViewScale[1] * scale, 0.25f, 8.0f);
    m_ViewScale[2] = MATH::CLAMP (m_ViewScale[2] * scale, 0.25f, 8.0f);
    rebuildMatrix ();
    rebound ();
    update ();
}

void Canvas::setXTranslation (int x) {
    m_ViewTranslate[0] = MATH::CLAMP_m11 ((float)x / width ());
    rebuildMatrix ();
    rebound ();
    repaint_constrained (rect ());
}

void Canvas::setYTranslation (int y) {
    m_ViewTranslate[1] = MATH::CLAMP_m11 ((float)y / height ());
    rebuildMatrix ();
    rebound ();
    repaint_constrained (rect ());
}

void Canvas::getTranslation (int & x, int & y) const {
    x = (int)(m_ViewTranslate[0] * width  ());
    y = (int)(m_ViewTranslate[1] * height ());
}

int Canvas::addXTranslation (int x) {
    if (x) {
        m_ViewTranslate[0] = MATH::CLAMP_m11 (m_ViewTranslate[0] + (float)x / width ());
        rebuildMatrix ();
        rebound ();
        repaint_constrained (rect ());
    }

    return (int) m_ViewTranslate[0];
}

int Canvas::addYTranslation (int y) {
    if (y) {
        m_ViewTranslate[1] = MATH::CLAMP_m11 (m_ViewTranslate[1] + (float)y / height ());
        rebuildMatrix ();
        rebound ();
        repaint_constrained (rect ());
    }

    return (int) m_ViewTranslate[1];
}

void Canvas::setRotation (float angle) {
    m_ViewRotate[2] = angle;
    while (m_ViewRotate[2] >= 360.0f) m_ViewRotate[2] -= 360.0f;
    while (m_ViewRotate[2] <  0.0f)   m_ViewRotate[2] += 360.0f;
    rebuildMatrix ();
    rebound ();
    update ();
}

void Canvas::addRotation (float angle) {
    m_ViewRotate[2] += angle;
    while (m_ViewRotate[2] >= 360.0f) m_ViewRotate[2] -= 360.0f;
    while (m_ViewRotate[2] <  0.0f)   m_ViewRotate[2] += 360.0f;
    rebuildMatrix ();
    rebound ();
    update ();
}

float Canvas::getRotation (void) const {
    return m_ViewRotate[2];
}

void Canvas::setDpi (int dpi, bool layers) {
    m_ImageDpi = dpi;
    if (layers) {
        foreach (ICanvasLayer * layer, m_Layers) {
            layer->setDpi (dpi);
        }
    }

    updateRenderingChain ();
    repaint_constrained (rect ());
}

void Canvas::rebuildMatrix (void) {
    const float theta = m_ViewRotate[2] * 3.14f / 180.0f;
    const float mrot[3 * 3] = {
        cosf (theta), -sinf (theta), 0.0f,
        sinf (theta),  cosf (theta), 0.0f,
        0.0f,          0.0f,         1.0f
    };
    const float mroti[3 * 3] = {
        cosf (theta), sinf (theta), 0.0f,
       -sinf (theta), cosf (theta), 0.0f,
        0.0f,         0.0f,         1.0f
    };

    const float sx = m_ViewScale[0];
    const float sy = m_ViewScale[1];
    const float sz = m_ViewScale[2];
    const float wsyncx = ((float) width  ()) / m_PlaneWidth;
    const float wsyncy = ((float) height ()) / m_PlaneHeight;

    const float msca[3 * 3] = {
        sx, 0.0f, 0.0f,
        0.0f, sy, 0.0f,
        0.0f, 0.0f, sz
    };
    const float mscai[3 * 3] = {
        1.0f / sx, 0.0f, 0.0f,
        0.0f, 1.0f / sy, 0.0f,
        0.0f, 0.0f, 1.0f / sz
    };

    const float mtra[3 * 3] = {
        1.0f,               0.0f,               0.0f,
        0.0f,               1.0f,               0.0f,
        m_ViewTranslate[0] * wsyncx,
       -m_ViewTranslate[1] * wsyncy,
        m_ViewTranslate[2]
    };
    const float mtrai[3 * 3] = {
        1.0f,                0.0f,                0.0f,
        0.0f,                1.0f,                0.0f,
       -m_ViewTranslate[0] * wsyncx,
       +m_ViewTranslate[1] * wsyncy,
       -m_ViewTranslate[2]
    };

    float m1[3 * 3];
    float m2[3 * 3];

    // Foreward transform matrix.
    g_MAT3x3_identity  ((float*)m1);
    g_MAT3x3_mmultiply ((const float*)m1, mtra, (float*)m2);
    g_MAT3x3_mmultiply ((const float*)m2, msca, (float*)m1);
    g_MAT3x3_mmultiply ((const float*)m1, mrot, (float*)m_matModelView);

    // Backward transform matrix.
    g_MAT3x3_identity  ((float*)m1);
    g_MAT3x3_mmultiply ((const float*)m1, mtrai, (float*)m2);
    g_MAT3x3_mmultiply ((const float*)m2, mscai, (float*)m1);
    g_MAT3x3_mmultiply ((const float*)m1, mroti, (float*)m_matModelView_inv);
}

void Canvas::rebound (void) {
}

ICanvasLayer * Canvas::allocateNewLayer (void) {
    if (m_Helper) {
        ICanvasLayer * layer = g_fac_SPAWN_CANVASLAYER (m_PlaneWidth, m_PlaneHeight, m_Helper);
        layer->setDpi (dpi ());
        return layer;
    } else
        return NULL;
}

void Canvas::append (ICanvasLayer * layer) {
    m_LastError = NULL;
    if (m_Layers.size () < MAX_CANVAS_LAYERS) {
        if (layer) {
            layer->addRef ();
            m_Layers.append (layer);

            if (m_CurrentLayer == NULL) {
                m_CurrentLayer = layer;
                logMessage (LOGIMPORTANCE_NEVERMIND, "New layer appended and set as current.");
            } else {
                logMessage (LOGIMPORTANCE_NEVERMIND, "New layer appended.");
            }

            updateRenderingChain ();
            m_Changed = TRUE;
        }
    } else {
        m_LastError = "Maximum layers count.";
        logMessage (LOGIMPORTANCE_ERROR, "Maximum layers count (%d) reached.", MAX_CANVAS_LAYERS);
    }
}

void Canvas::prepend (ICanvasLayer * layer) {
    m_LastError = NULL;
    if (m_Layers.size () < MAX_CANVAS_LAYERS) {
        if (layer) {
            layer->addRef ();
            m_Layers.prepend (layer);

            if (m_CurrentLayer == NULL) {
                m_CurrentLayer = layer;
                logMessage (LOGIMPORTANCE_NEVERMIND, "New layer prepended and set as current.");
            } else {
                logMessage (LOGIMPORTANCE_NEVERMIND, "New layer prepended.");
            }

            updateRenderingChain ();
            m_Changed = TRUE;
        }
    } else {
        m_LastError = g_err_MAXLAYERS;
        logMessage (LOGIMPORTANCE_ERROR, "Maximum layers count (%d) reached.", MAX_CANVAS_LAYERS);
    }
}

void Canvas::removeAt (int index) {
    m_LastError = NULL;
    if (index >= 0 && index < m_Layers.size ()) {
        ICanvasLayer * layer = m_Layers.at (index);

        // Shift the selection.
        if (layer == m_CurrentLayer) {
            if (index < m_Layers.size () - 1)
                m_CurrentLayer = m_Layers.at (index + 1);
            else if (index > 0)
                m_CurrentLayer = m_Layers.at (index - 1);
            else
                m_CurrentLayer = NULL;
        }

        layer->release ();
        m_Layers.removeAt (index);
        logMessage (LOGIMPORTANCE_NEVERMIND, "Layer at position %d has been removed.", index);
        m_Changed = TRUE;
    } else {
        m_LastError = g_err_WRONGIDX;
        logMessage (LOGIMPORTANCE_ERROR, m_LastError);
    }
}

ICanvasLayer * Canvas::at (int index) {
    return m_Layers.at (index);
}

void Canvas::clear (void) {
    if (m_Layers.size ()) {
        reset ();
        updateRenderingChain ();
    }
}

void Canvas::erase (void) {
    makeCurrent ();

    m_LastError = NULL;
    m_Changed   = TRUE;
    foreach (ICanvasLayer * layer, m_Layers) {
        layer->clear ();
    }

    update ();
}

void Canvas::reset (void) {
    m_LastError = NULL;
    foreach (ICanvasLayer * layer, m_Layers) {
        layer->release ();
    }

    m_Layers.clear ();
    m_Changed      = TRUE;
    m_CurrentLayer = NULL;
    updateRenderingChain ();
    update ();

    logMessage (LOGIMPORTANCE_NEVERMIND, "Canvas has been reset.");
}

bool Canvas::applyPipeline (IPipeline * pipe, float dt) {
    // Skip the implementation if:
    // - we have no context yet ;
    // - we are in stroke now ;
    // - we are in drag now.
    if (!m_HasGLContext  ||
         m_StrokeStarted ||
         m_DragMode != DRAGSTATE_NOOP)
        return FALSE;

    makeCurrent ();
    bool changed = FALSE;
    foreach (ICanvasLayer * layer, m_Layers) {
        if (layer) {
            // Attach all channels to pipeline.
            for (int i = 0; i < layer->size (); i++) {
                IChannel * channel = layer->at (i);
                if (channel) {
                    channel->pokeObject ();
                    pipe->pushChannel (channel);
                }
            }

            // And execute it.
            changed |= pipe->iterate (dt);
       }
    }

    m_Changed |= changed;
    if (changed) {
        repaint_constrained (rect ());
    }

    return changed;
}

void Canvas::statusChanged (QHASGlobal::eStatus status) {
    m_GlobalStatus = status;
    update ();
}

bool Canvas::promoteChannel (FOURCC uid) {
    makeCurrent ();

    bool added = FALSE;
    m_LastError = NULL;
    if (m_CurrentLayer && m_HasGLContext) {
        // Check if the promoted channel
        // exists in current layer.
        if (!m_CurrentLayer->hasChannel (uid)) {
            int w = m_PlaneWidth;
            int h = m_PlaneHeight;

            // If it doesn't, add it.
            IChannel * channel = g_fac_SPAWN_CHANNEL (uid, w, h, m_Helper);
            if (channel) {
                logMessage (LOGIMPORTANCE_NEVERMIND, "Added promoted channel: %s.", IChannel::nameFromUID (uid));
                m_CurrentLayer->append (channel);
                channel->clear ();
                channel->release ();
                added = TRUE;
            } else {
                logMessage (LOGIMPORTANCE_ERROR, "Failed to create promoted channel: %s.", IChannel::nameFromUID (uid));
            }
        }
    }

    // And check the mixer channels
    // if it is visible.
    /*
    if (m_MixerVisible &&
        m_MixerLayer && (
        CHANNELUID_RaGaBaW_dynamic == uid ||
        CHANNELUID_RsGsBsW_dynamic == uid)) {
        if (!m_MixerLayer->hasChannel (uid)) {
            IChannel * channel = g_fac_SPAWN_CHANNEL (uid, m_MixerWidth, m_MixerHeight, m_Helper);
            if (channel) {
                logMessage (LOGIMPORTANCE_NEVERMIND, "Added promoted channel to mixer: %s.", IChannel::nameFromUID (uid));
                m_MixerLayer->append (channel);
                channel->release ();
                added = TRUE;
            } else {
                logMessage (LOGIMPORTANCE_ERROR, "Failed to create promoted channel: %s.", IChannel::nameFromUID (uid));
            }
        }
    }
    */

    if (added) {
        updateRenderingChain ();
        update ();
    }

    return added;
}

void Canvas::setMixerVisibility (bool visible) {
    m_MixerVisible = visible;
    updateRenderingChain ();
    update ();
}

void Canvas::applyFilter (IFilter * filter) {
    m_LastError = NULL;
    if (filter) {
        // FIXME Not implemented;
        m_Changed = TRUE;
    }
}

void Canvas::initializeGL (void) {
    logMessage (LOGIMPORTANCE_NEVERMIND, "My life for Aiur!..");

    if (doubleBuffer ()) {
        logMessage (LOGIMPORTANCE_NEVERMIND, "Good. I'll use double buffering here.");
    } else {
        logMessage (LOGIMPORTANCE_NEVERMIND, "Double buffering is disabled.");
    }

    if (m_Helper) {
        m_Helper->init ();
        m_HasGLContext = m_Helper->inited ();
        if (m_HasGLContext) {
          //m_MixerLayer = g_fac_SPAWN_CANVASLAYER (m_MixerWidth, m_MixerHeight, m_Helper);
            updateRenderingChain ();

            emit glcontextObtained ();
        }
    }

    if (m_HasGLContext)
        logMessage (LOGIMPORTANCE_NEVERMIND, "Packed up and good to go!");
    else {
        logMessage (LOGIMPORTANCE_ERROR, "Something went wrong... staying uninitialized.");
    }
}

void Canvas::resizeGL (int width, int height) {
    Q_UNUSED (width);
    Q_UNUSED (height);

    // FIXME Maybe unneeded.
    rebuildMatrix ();
}

void Canvas::render (const QRect & rect) {
    glMatrixMode (GL_MODELVIEW);
    glPushMatrix ();
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();

    if (m_RenderingChain && m_CurrentLayer) {
        const bool empty = !m_CurrentLayer->hasChannel (CHANNELUID_Optical);

        if (!m_HasGLContext || empty) {
            // Paint a blank screen.
            // FIXME Maybe paint something different. =)
            glClearColor (0.4f, 0.4f, 0.45f, 1.0f);
            glClear      (GL_COLOR_BUFFER_BIT);
        } else {
            // Describe the operation.
            sRenderOperation operation;
            operation.target_w   = width ();
            operation.target_h   = height ();
            operation.target_dpi = dpi ();
            operation.roi_x      = rect.left ();
            operation.roi_y      = rect.top ();
            operation.roi_w      = rect.width  ();
            operation.roi_h      = rect.height ();

            operation.plane_width  = m_PlaneWidth;
            operation.plane_height = m_PlaneHeight;

            // Canvas plane transformation.
            operation.view_translate[0] = m_ViewTranslate[0];
            operation.view_translate[1] = m_ViewTranslate[1];
            operation.view_translate[2] = m_ViewTranslate[2];
            operation.view_scale[0]     = m_ViewScale[0] * m_PlaneWidth  / width ();
            operation.view_scale[1]     = m_ViewScale[1] * m_PlaneHeight / height ();
            operation.view_scale[2]     = 1.0f;
            operation.view_rotate[0]    = 0.0f;
            operation.view_rotate[1]    = 0.0f;
            operation.view_rotate[2]    = m_ViewRotate[2];

            // Restore the default framebuffer.
            glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);

            // Set the rendered thickness of a paint layer.
            if (m_PaintingStyle) {
                m_RenderingChain->setPaintThickness (
                    1.0f * m_PaintingStyle->paintViscosity ()
                );
            }

            // Render on a (back)buffer.
            m_RenderingChain->render (
                doubleBuffer () ? GL_BACK : GL_FRONT,
                &operation
            );
        }
    } else {
        // FIXME Paint a blank screen.
        glClearColor (0.4f, 0.4f, 0.45f, 1.0f);
        glClear      (GL_COLOR_BUFFER_BIT);

        // Request a rendering chain for we need one.
        if (m_Connector && m_HasGLContext) {
            m_ReqMutex.lock ();
            bool masked = m_ReqMap[RQ_RENDERINGCHAIN];
            m_ReqMutex.unlock ();

            if (!masked) {
                m_ReqMutex.lock ();
                m_ReqMap[RQ_RENDERINGCHAIN] = TRUE;
                m_ReqMutex.unlock ();
                m_Connector->allocateObject (this, RQ_RENDERINGCHAIN, ENTITY_RENDERING_CHAIN);
            }
        }
    }

    glMatrixMode (GL_PROJECTION);
    glPopMatrix ();
    glMatrixMode (GL_MODELVIEW);
    glPopMatrix ();
}

//
void Canvas::paintEvent (QPaintEvent * event) {
    // Sometimes it is needed. x_x'
    makeCurrent ();
    //glMatrixMode (GL_MODELVIEW);
    //glPushMatrix ();

    // First, render the canvas plane.
    render (event->rect ());

    //glMatrixMode (GL_MODELVIEW);
    //glPopMatrix ();

    // Restore the framebuffer.
    // glBindFramebufferEXT (GL_FRAMEBUFFER_EXT, 0);

    // Second, overpaint the widget with
    // some graphics.
    QPainter painter (this);
    if (m_Instrument && m_LastPosInside_w) {
        painter.setPen (QPen (QColor (0, 0, 0)));
        painter.drawEllipse (m_rPointer);
    }

    // Third, check if we have an active
    // selection and paint it.
    if (m_rSelection.width () *
        m_rSelection.height ()) {

        const int cx = m_PlaneWidth / 2;
        const int cy = m_PlaneHeight / 2;

        const float corner0[3] = {
            (float)(m_rSelection.left() - cx) / m_PlaneWidth,
            (float)(m_rSelection.top()  - cy) / m_PlaneHeight, 1.0f};
        const float corner1[3] = {
            (float)(m_rSelection.right()  - cx) / m_PlaneWidth,
            (float)(m_rSelection.bottom() - cy) / m_PlaneHeight, 1.0f};

        float corner2[3];
        float corner3[3];

        // Transform the selection into the window space.
        g_MAT3x3_vmultiply (corner0, (float*)m_matModelView, corner2);
        g_MAT3x3_vmultiply (corner1, (float*)m_matModelView, corner3);
        corner2[0] = corner2[0] * m_PlaneWidth  + width () / 2;
        corner3[0] = corner3[0] * m_PlaneWidth  + width () / 2;
        corner2[1] = corner2[1] * m_PlaneHeight + height () / 2;
        corner3[1] = corner3[1] * m_PlaneHeight + height () / 2;

        // Draw the selection.
        painter.setPen (QPen (QColor (0, 0, 0), 1, Qt::DotLine));
        painter.drawRect (QRect (
                (int)(corner2[0]),
                (int)(corner2[1]),
                (int)(corner3[0] - corner2[0]),
                (int)(corner3[1] - corner2[1])
        ));
    }

    painter.end ();
}
//
/*
void Canvas::paintGL (void) {
    render (rect ());
}
*/

void Canvas::repaint_constrained (const QRect & rect) {
    if (m_CumulativeRectTrigger) {
        m_rCumulative = m_rCumulative.united (rect);
    } else {
        m_rCumulative = rect;
        m_CumulativeRectTrigger = TRUE;
    }

    // Constrain the repainting rate.
    if (m_timerUpdate.elapsed () > 1000 / FPS_LIMIT) {
        int R;
        if (m_Instrument) {
            R = (int)(0.5f * getUniformScale () * m_Instrument->currentPrimitiveWidth ());
            R = MAX (4, R);
        } else
            R = 0;

        QRect rupdate  = m_rPointer;
        m_rPointer = QRect (m_LastXPos_w - R, m_LastYPos_w - R, 2 * R, 2 * R);

        // Compose the updating region of:
        // * target region ;
        // * old cursor position ;
        // * new cursor position.
        rupdate = rupdate.united (m_rPointer);
        QRect rupdate1 = rupdate;
        rupdate1.setTop    (height () - rupdate.bottom ());
        rupdate1.setBottom (height () - rupdate.top ());
        rupdate = m_rCumulative;
        m_rCumulative.setTop    (height () - rupdate.bottom ());
        m_rCumulative.setBottom (height () - rupdate.top ());
        rupdate1 = rupdate1.united (m_rCumulative).adjusted (-8, -8, +8, +8);

        repaint (rupdate1);
        m_CumulativeRectTrigger = FALSE;

        m_timerUpdate.restart ();
    }
}

void Canvas::poke (void) {
    if (!m_StrokeStarted) {
        m_TabletPresence = FALSE;
    }
}

/// Receive the data from server.
void Canvas::complitedRequest (int param, int length, void * data) {
    Q_UNUSED (length);

    bool error = FALSE;
    if (param == RQ_RENDERINGCHAIN) {
        SAFE_RELEASE (m_RenderingChain);

        // Save the object pointer.
        m_RenderingChain = static_cast<IRenderingChain*>(data);
        if (m_RenderingChain && m_HasGLContext) {
            updateRenderingChain ();

            emit renderingOptsRequest ();
        }

        // Repaint the widget.
        update ();
    } else
    if (param == RQ_PIPELINES) {
        IPipeline * pipe = static_cast<IPipeline*>(data);

        // Check the pipeline class.
        if (0 == strcmp (pipe->myClassName (), "Server::Pipeline_stage0")) {
            SAFE_RELEASE (m_PipelineDry);

            // Save the object pointer.
            m_PipelineDry = pipe;
            logMessage (LOGIMPORTANCE_NEVERMIND, "Recieved the pipelines.");
        } else
            pipe->release ();
    } else {
        logMessage (LOGIMPORTANCE_ERROR, "Undefined server message: %d", param);
        error = TRUE;
    }

    if (!error) {
        // Unmask the request.
        m_ReqMutex.lock ();
        m_ReqMap[param] = FALSE;
        m_ReqMutex.unlock ();
    }
}

ISpline * Canvas::rebuildSpline_fbmu (float rsz0) {
    // Reduce the reservoir effect for the viscous paints.
    // FIXME Needs adjustment.
    const float rsz = rsz0 * (1.00f - 0.25f * m_PaintingStyle->paintViscosity ());

    if (m_SplineFeedbackMu == NULL ||
        IS_SIGNIFICANT (fabsf (m_LastReservoirSz - rsz))) {
        SAFE_RELEASE (m_SplineFeedbackMu);

        const double spline_len[3] = {
            0.00,
            100.0 * rsz,
            150.0
        };

        const double spline_mul[3] = {
            0.0, 0.3, 1.0
        };

        m_SplineFeedbackMu = QHASMath::newSpline (spline_len, spline_mul, 3);
        m_LastReservoirSz = rsz;
    }

    return m_SplineFeedbackMu;
}

void Canvas::mousePressEvent (QMouseEvent * event) {
    if (!m_TabletPresence) {
        if (m_MouseHash.contains (event->button())) {
            setDragMode (m_MouseHash.value (event->button()));
        } else
            setDragMode (DRAGSTATE_PAINT);

        if (m_DragMode == DRAGSTATE_PAINT) {
            if (m_Instrument) {
                if (m_UseSmartStroke &&
                    m_Instrument->instrumentGroup () == INSTRUMENTGROUP_BRUSH) {
                    m_SmartStroke_mu = 0.0f;
                } else {
                    m_Instrument->setPressure (0.8f, INPUT_CONTROLLER_MOUSE);
                    m_SmartStroke_mu = 0.8f;
                }

                m_Instrument->meterReset ();
                m_Instrument->anchoredMeterReset ();
            } else {
                m_SmartStroke_mu = 0.8f;
            }

            // Mark that user is painting.
            if (!m_StrokeStarted) {
                emit userStartsPainting ();
            }
        }

        // And call the handler.
        mouseMoveEvent (event);
    }
}

void Canvas::mouseReleaseEvent (QMouseEvent * event) {
    Q_UNUSED (event);
    if (!m_TabletPresence) {
        // Reset the modifiers state.
        setDragMode (DRAGSTATE_PAINT);

        if (m_StrokeStarted) {
            finishStroke ();
        }

        if (m_Instrument) {
            m_Instrument->setPressure (0.0f, INPUT_CONTROLLER_MOUSE);
            m_Instrument->reset ();
        }

        update ();
    }
}

void Canvas::enterEvent (QEvent * event) {
    Q_UNUSED (event);

    emit gainedFocus ();
}

void Canvas::leaveEvent (QEvent * event) {
    Q_UNUSED (event);

    if (m_StrokeStarted) {
        finishStroke ();
    }

    if (m_Instrument) {
        m_Instrument->setPressure (0.0f, INPUT_CONTROLLER_MOUSE);
        m_Instrument->reset ();
    }

    m_LastPosInside_w = FALSE;
    update ();
}

void Canvas::finishStroke (void) {
    // Clear the patches cache.
    if (m_Helper &&
        m_HasGLContext) {
        for (int x = 0; x < 64; x++) {
            for (int y = 0; y < 64; y++) {
                for (int i = 0; i < 9; i++) {
                    unsigned long key = g_MAKE_FOURCC (x, y, i, 1);
                    if (m_Helper->sharedPatchCache ()->contains (key)) {
                        m_Helper->sharedPatchUnlock (
                            m_Helper->sharedPatchCache ()->take (key)
                        );
                    }
                }
            }
        }
    }

    m_StrokeStarted = FALSE;

    m_LineStrokeLength[0]   = 0;
    m_LineStrokeLength[1]   = 0;
    m_CurrentStrokeLength_x = 0;
    m_CurrentStrokeLength_y = 0;

    // Finish the transaction if any.
    if (m_CurrentTransaction) {
        m_CurrentTransaction->commit ();
        m_CurrentTransaction = NULL;
    }

    emit userStopsPainting ();
}

void Canvas::mouseMoveEvent (QMouseEvent * event) {
	PROFILE_FUNC();
    if (!m_TabletPresence) {
        if (m_LastDevUid != 0) {
            m_LastDevUid  = 0;
            emit tabletDeviceChanged (0);
        }

        instrumentMoves (
            event->x (),
            event->y (),
            event->buttons() & Qt::LeftButton
        );

        m_LastXPos_w = event->x ();
        m_LastYPos_w = event->y ();
    }
}

void Canvas::tabletEvent (QTabletEvent * event) {
    qint64 uid = event->uniqueId ();

    if (m_LastDevUid != uid) {
        m_LastDevUid  = uid;
        emit tabletDeviceChanged (uid);
    }

    // Check if this device is new to us.
    if (!m_Devices.contains (uid)) {
         m_Devices.append (uid);

#ifdef DEBUG_INPUTDEVICE_INFORMATION
        switch (event->device ()) {
            case QTabletEvent::Puck:
                logMessage (LOGIMPORTANCE_NEVERMIND, "I feel the presence of a Puck-device (uid: %d).", (int) uid);
                break;
            case QTabletEvent::Stylus:
                logMessage (LOGIMPORTANCE_NEVERMIND, "I feel the presence of a tablet device with Stylus (uid: %d).", (int) uid);
                break;
            case QTabletEvent::Airbrush:
                logMessage (LOGIMPORTANCE_NEVERMIND, "I feel the presence of a tablet device with Airbrush (uid: %d).", (int) uid);
                break;
            case QTabletEvent::FourDMouse:
                logMessage (LOGIMPORTANCE_NEVERMIND, "I feel the presence of a 4D mouse (uid: %d).", (int) uid);
                break;
            case QTabletEvent::RotationStylus:
                logMessage (LOGIMPORTANCE_NEVERMIND, "I feel the presence of a tablet device with 6D stylus (uid: %d). O_o'", (int) uid);
                break;
            default:
                logMessage (LOGIMPORTANCE_NEVERMIND, "I feel the presence some unrecognized tablet device (uid: %d). x_x", (int) uid);
                break;
        }
#endif
    }

    if (m_LastPointer != event->pointerType ()) {
        // Notify the world that pen has been changed.
        emit tabletPenChanged (event->pointerType ());

#ifdef DEBUG_INPUTDEVICE_INFORMATION
        switch (event->pointerType ()) {
            case QTabletEvent::Pen:
                logMessage (LOGIMPORTANCE_NEVERMIND, "Tablet device has been set to Pen.");
                break;
            case QTabletEvent::Cursor:
                logMessage (LOGIMPORTANCE_NEVERMIND, "Tablet device has been set to Cursor.");
                break;
            case QTabletEvent::Eraser:
                logMessage (LOGIMPORTANCE_NEVERMIND, "Tablet device has been set to Eraser.");
                break;
            default:
                logMessage (LOGIMPORTANCE_NEVERMIND, "Tablet device has been set to unknown pointer type.");
                break;
        }
#endif
        m_LastPointer = event->pointerType ();
    }

    switch (event->pointerType ()) {
        // Accept all pointer types now.
        case QTabletEvent::Pen:
        case QTabletEvent::Cursor:
        case QTabletEvent::Eraser:
            m_LastXPos_w = event->x ();
            m_LastYPos_w = event->y ();

            m_TabletPresence = TRUE;
            if (IS_SIGNIFICANT (event->pressure ())) {
			  // don't start paint while special modifier is active
			  if ((m_DragMode == DRAGSTATE_NOOP || m_DragMode == DRAGSTATE_PAINT)
				  && !m_StrokeStarted) {
                     emit userStartsPainting ();

                     // FIXME Unfinished.
                     setDragMode (DRAGSTATE_PAINT);
                }

                if (m_Instrument) {
                    if (IS_SIGNIFICANT (m_PressureWRng[1] - m_PressureWRng[0])) {
                        // Map the pressure into a working range.
                        const
                        float pressure = ((float) event->pressure () - m_PressureWRng[0]) /
                                                  (m_PressureWRng[1] - m_PressureWRng[0]);

                        // Pass the pressure to instrument.
                        m_Instrument->setPressure (
                                pressure,
                                INPUT_CONTROLLER_TABLET
                        );
                    }

                    // Pass the tilt to instrument.
                    // FIXME Tilt is disabled.
                    /*
                    m_Instrument->setTilt (
                            event->xTilt (),
                            event->yTilt (),
                            INPUT_CONTROLLER_TABLET
                    );
                    */

                    m_Instrument->meterReset ();
                    m_Instrument->anchoredMeterReset ();
                }

                instrumentMoves (event->x (), event->y (), TRUE);
                event->accept ();
                return;
            } else {
                instrumentMoves (event->x (), event->y (), FALSE);
                if (m_StrokeStarted) {
                    leaveEvent (event);
                }
            }
        break;
        default:
        break;
    }

    event->accept ();
}

void Canvas::instrumentMoves (int x0, int y0, bool pressed) {
    PROFILE_FUNC();

    if (!m_Instrument)
        return;

    const float cx0 = 0.5f * width ();
    const float cy0 = 0.5f * height ();
    const int   cx  = m_PlaneWidth / 2;
    const int   cy  = m_PlaneHeight / 2;

    const float pos0[] = {
        ((float)x0 - cx0) / m_PlaneWidth,
        ((float)y0 - cy0) / m_PlaneHeight,
        1.0f
    };
    float pos[3];

    // Translate the point to canvas coordinates.
    g_MAT3x3_vmultiply (pos0, (float*) m_matModelView_inv, pos);

    const float fx = m_PlaneWidth  * pos[0] + cx;
    const float fy = m_PlaneHeight * pos[1] + cy;

    const int x  = (int) (fx);
    const int y  = (int) (fy);
    const int dx = x - m_LastXPos;
    const int dy = y - m_LastYPos;
    const int T = MAX (2, m_Instrument->currentPrimitiveWidth () / 32);
    if (abs (dx) < T &&
        abs (dy) < T &&
        m_StrokeStarted) {
        return;
    }

    bool inside = TRUE;
    if (x < 0 || x > m_PlaneWidth ||
        y < 0 || y > m_PlaneHeight)
        inside = FALSE;

    m_LastPosInside_w = inside;

    // Broadcast the current pointer position.
    emit tracking (fx, fy,
                   m_Instrument->velocityX (),
                   m_Instrument->velocityY (),
                   m_Instrument->pressure (),
                   inside
    );

    // Process the dragging event if any.
    // Should not demand the mouse to be pressed..
    if (/*pressed && */m_DragMode == DRAGSTATE_CANVAS_MOVE) {
        addXTranslation (x0 - m_LastXPos_w);
        addYTranslation (m_LastYPos_w - y0);
    } else
    if (/*pressed && */m_DragMode == DRAGSTATE_CANVAS_ROTATE) {
        addRotation ((float)(m_LastYPos_w - y0));
    } else
    if (/*pressed && */m_DragMode == DRAGSTATE_CANVAS_SCALE) {
        multiplyScale (MATH::CLAMP (1.0f + 0.01f * (m_LastYPos_w - y0), .0f, 2.f));
    } else {
        if (!pressed || !m_HasGLContext) {
            m_LastXPos = x;
            m_LastYPos = y;

            if (m_HasGLContext) {
                m_Instrument->setPosition (x, y,
                                           m_TabletPresence ? INPUT_CONTROLLER_TABLET :
                                                              INPUT_CONTROLLER_MOUSE);
                // Instrument early init.
                m_Instrument->setPressure (0.0f, INPUT_CONTROLLER_GOD);
            }

            // I pass the empty rect hoping that it'll do the trick.
            repaint_constrained (QRect ());
            return;
        }

        // Select the target layer.
        ICanvasLayer * target_layer = m_CurrentLayer;

        // Check if user hit the mixer layer.
        if (m_MixerVisible &&
            m_MixerLayer) {
            if (x0 > 0 &&
                x0 < 0 + m_MixerWidth &&
                y0 > 0 &&
                y0 < 0 + m_MixerHeight) {
                target_layer = m_MixerLayer;
            }
        }

        // Apply an instrument if any.
        if (inside && m_DragMode == DRAGSTATE_PAINT) {
            if (m_Instrument) {
                for (int i = 0; i < 1 + !m_StrokeStarted; i++) {
                    m_Instrument->setPosition (x, y,
                                               m_TabletPresence ? INPUT_CONTROLLER_TABLET :
                                                                  INPUT_CONTROLLER_MOUSE);
                }

                QRect rInvalidated = applyInstrument (m_Instrument, target_layer);
                m_StrokeStarted = TRUE;

                if (rInvalidated.width  () &&
                    rInvalidated.height ()) {
                    const float pos0[] = {
                        (float)(rInvalidated.left () - cx) / m_PlaneWidth,
                        (float)(rInvalidated.top ()  - cy) / m_PlaneHeight,
                        1.0f
                    };
                    const float pos1[] = {
                        (float)(rInvalidated.right ()  - cx) / m_PlaneWidth,
                        (float)(rInvalidated.bottom () - cy) / m_PlaneHeight,
                        1.0f
                    };
                    float pos2[3];
                    float pos3[3];

                    // Translate back into the widget space.
                    g_MAT3x3_vmultiply (pos0, (float*) m_matModelView, pos2);
                    g_MAT3x3_vmultiply (pos1, (float*) m_matModelView, pos3);

                    const int margin = 5;
                    repaint_constrained (QRect (
                        (int)(+pos2[0] * m_PlaneWidth  + cx0      -     margin),
                        (int)(-pos2[1] * m_PlaneHeight + cy0      -     margin),
                        (int)((pos3[0] - pos0[0]) * m_PlaneWidth  + 2 * margin),
                        (int)((pos3[1] - pos0[1]) * m_PlaneHeight + 2 * margin)
                    ));

                    // Check the stroke direction.
                    if (m_LineStrokeLength[0] > 16 ||
                        m_LineStrokeLength[1] > 16) {
                        if ((m_StrokeDirection[0] * m_Instrument->velocityX () < 0.0f &&
                             m_LineStrokeLength[0] > 16) ||
                            (m_StrokeDirection[1] * m_Instrument->velocityY () < 0.0f &&
                             m_LineStrokeLength[1] > 16)) {

                            m_StrokeDirection[0] = m_Instrument->velocityX ();
                            m_StrokeDirection[1] = m_Instrument->velocityY ();
                            m_LineStrokeLength[0] = 0;
                            m_LineStrokeLength[1] = 0;

                            emit strokeTurns ();
                        }
                    } else {
                        m_StrokeDirection[0] = m_Instrument->velocityX ();
                        m_StrokeDirection[1] = m_Instrument->velocityY ();
                    }

                    m_LineStrokeLength[0] += abs (dx);
                    m_LineStrokeLength[0] += abs (dy);
                }
            }
        }

        // Increase the stroke length.
        m_CurrentStrokeLength_x += abs (dx);
        m_CurrentStrokeLength_y += abs (dy);

        m_LastXPos = x;
        m_LastYPos = y;
    }
}

void Canvas::setInstrument (IInstrumentProxy * instrument) {
    SAFE_RELEASE (m_Instrument);
    if (instrument) {
        m_Instrument = instrument;
        m_Instrument->addRef ();
        if (m_Instrument && m_HasGLContext) {
            if (m_Instrument->instrumentGroup () != INSTRUMENTGROUP_TOOLBOX) {
                m_Instrument->setPressure (0.0f, INPUT_CONTROLLER_MOUSE);
            }
        }
    }
}

/// Use the idle-time to refilter the paper texture
/// with a new instrument size.
/// FIXME Deprecated.
void Canvas::instrumentSizeChanged (IInstrumentProxy * instrument,
                                    int radius) {
}

void Canvas::setPaperTexture (sPaper * texture) {
    m_Papermap->setPaper (texture);
    updateRenderingChain ();
    update ();
}

void Canvas::layerChanged (ICanvasLayer * layer) {
    m_CurrentLayer = layer;
}

/// Apply the instrument to current layer.
QRect Canvas::applyInstrument (IInstrumentProxy * instrument, ICanvasLayer * layer) {
    QRect rInvalidated = QRect ();
	if(!instrumentMutex.tryLock()) return rInvalidated;
    PROFILE_FUNC();

    if (!m_HasGLContext) {
        m_LastError = g_err_NOCONTEXT;
    }

    bool  invalidated_trigger = FALSE;

    m_LastError = NULL;

    makeCurrent ();
    if (instrument && layer) {

        // Save the instrument state.
        for (int i = 0; i < layer->size (); i++) {
            instrument->statePush ();
        }

        float smartStroke_mu_bk = m_SmartStroke_mu;

        // Walk through all channels.
        for (int i = 0; i < layer->size (); i++) {
            IChannel * channel = layer->at (i);
            m_CurrentChannel = channel;

            // Restore the instrument state.
            instrument->statePop ();
            m_SmartStroke_mu = smartStroke_mu_bk;

            int strokex0, strokex1;
            int strokey0, strokey1;
            instrument->stroke (0.0f, strokex0, strokey0);
            instrument->stroke (1.0f, strokex1, strokey1);

            const int R     = instrument->currentPrimitiveWidth ();
            const int span  = 1 + MAX (abs (strokex1 - strokex0), abs (strokey1 - strokey0));
            const int step  = MAX (1, R / STROKE_ISTEP_R());

            m_Stroke_len = 0;

            if (strokex0 != strokex1 ||
                strokey0 != strokey1 ||
                !m_StrokeStarted) {
                // Generate the stroke from interpolated points.
                for (int j = 0; j < span && m_Stroke_len < 1024; j += step) {
                    instrument->stroke (
                        (float)j / span,
                        strokex0,
                        strokey0
                    );

                    if (abs (strokex0 - strokex1) > 1 ||
                        abs (strokey0 - strokey1) > 1 ||
                        j == 0) {
                        strokex1                 = strokex0;
                        strokey1                 = strokey0;
                        m_Stroke[m_Stroke_len].x = strokex0;
                        m_Stroke[m_Stroke_len].y = strokey0;

                        m_Stroke[m_Stroke_len].stroke_length = m_CurrentStrokeLength_x +
                                                               m_CurrentStrokeLength_y +
                                                               m_Stroke_len;

                        m_Stroke[m_Stroke_len].feedback = FALSE;
                        m_Stroke_len++;

                        if (invalidated_trigger) {
                            rInvalidated = rInvalidated.united (
                                QRect (strokex0 - R, strokey0 + R, 2 * R, 2 * R)
                            );
                        } else {
                            rInvalidated = QRect (strokex0 - R, strokey0 + R, 2 * R, 2 * R);
                            invalidated_trigger = TRUE;
                        }
                    }
                }
            }

            // Apply the brush at each point.
            for (int k = 0; k < m_Stroke_len; k++) {
                // Brush pressure multiplier for a mouse.
                if (!m_TabletPresence) {
                    if (m_UseSmartStroke &&
                        instrument->instrumentGroup () == INSTRUMENTGROUP_BRUSH) {
                        instrument->setPressure (m_SmartStroke_mu, INPUT_CONTROLLER_MOUSE);
                        m_SmartStroke_mu = MIN (0.8f, m_SmartStroke_mu + 0.04f);
                    } else {
                        instrument->setPressure (0.8f, INPUT_CONTROLLER_MOUSE);
                    }
                }

                // Apply the Instrument.
                m_Stroke_idx = k;
                instrument->applyInstrument (channel, dynamic_cast<ICanvasDelegate*>(this));
            }
        }

        m_Changed = TRUE;
    }

	instrumentMutex.unlock();
    return rInvalidated;
}

void Canvas::touch (IInstrumentLock * operation) {
    IInstrumentProxy * instrument = m_Instrument;
    if (operation && instrument) {
        IChannel * channel = m_CurrentChannel;
        if (!channel)
            return;

        // Start a new transtaction if not started yet.
        if (!m_CurrentTransaction) {
            if (m_ActionControl) {
                m_CurrentTransaction = m_ActionControl->beginTransaction ();
            }
        }

        // Build the extended operation descriptor.
        CanvasOperation opex;

        opex.m_Velocity[0]          = instrument->velocityX ();
        opex.m_Velocity[1]          = instrument->velocityY ();
        opex.m_PaintingStyle        = m_PaintingStyle;
        opex.m_Source               = operation;
        opex.m_Instrument           = instrument;
        opex.m_Paper                = m_Papermap->detail (0.5f * instrument->currentPrimitiveWidth ());
        opex.m_PaperDpi             = m_Papermap->paperDpi ();
        opex.m_Transaction          = m_CurrentTransaction;

        ISpline * spline = rebuildSpline_fbmu (operation->reservoirSize ());
        if (spline) {
            if (m_Stroke[m_Stroke_idx].stroke_length < 150)
                opex.m_FeedbackMu = MATH::CLAMP_01 (spline->f (m_Stroke[m_Stroke_idx].stroke_length));
            else
                opex.m_FeedbackMu = 1.0f;
        }

        // Number of steps to skip due to
        // the low presition of a fragment program.
        const int T = MAX (1, (int)(0.05f / opex.suckingRate ()));
        if (instrument->meter () >= T &&
            m_StrokeStarted) {
            m_Stroke[m_Stroke_idx].feedback = TRUE;
            instrument->meterReset ();
        } else
            m_Stroke[m_Stroke_idx].feedback = FALSE;

        channel->touch (&opex, m_Stroke + m_Stroke_idx, 1);

        // Finally, release the operation descriptor.
        operation->release ();
    }
}

/// SLOW. Used only to save
/// the canvas contents to a file.
IArray * Canvas::asImage (void) {
    m_LastError = NULL;

    // Create a buffer for a saved picture.
    IArray * i = g_fac_SPAWN_ARRAY (m_Helper, m_PlaneWidth, m_PlaneHeight,
                                    GL_RGBA8, GL_UNSIGNED_BYTE, IArray::FLAGS_NULL);
    if (i) {
        logMessage (LOGIMPORTANCE_NEVERMIND, "Rendering on an auxilary pixel buffer...");
        return asImage (i);
    } else {
        m_LastError = "IArray creation error.";
        logMessage (LOGIMPORTANCE_CRITICAL, m_LastError);
        return NULL;
    }
}

/// The same as above, but the target is arbitrary.
/// Used to create the canvas preview.
/// @return Returns target.
IArray * Canvas::asImage (IArray * target) {
    m_LastError = NULL;
    setMixerVisibility (FALSE);

    if (m_RenderingChain) {
        if (!m_HasGLContext) {
            m_LastError = g_err_NOCONTEXT;
            logMessage (LOGIMPORTANCE_CRITICAL, m_LastError);
            return NULL;
        }

        makeCurrent ();
        IFBOWrapper * fbo = m_Helper->sharedFBO (OPENGLHELPER_FBO_GENERIC);
        if (fbo) {
            GLuint tex = target->lock_gltex ();
            if (tex != UINT_UNDEFINED) {
                fbo->bind ();
                glFramebufferTexture2DEXT (GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                                           GL_TEXTURE_2D, tex, 0);

                // Describe the operation.
                sRenderOperation operation;
                memset (&operation, 0, sizeof(sRenderOperation));

                operation.plane_width   = m_PlaneWidth;
                operation.plane_height  = m_PlaneHeight;
                operation.target_w      = target->width ();
                operation.target_h      = target->height ();
                operation.target_dpi    = dpi ();
                operation.roi_x         = 0;
                operation.roi_y         = 0;
                operation.roi_w         = target->width ();
                operation.roi_h         = target->height ();
                operation.view_scale[0] = +1.0f;
                operation.view_scale[1] = -1.0f;
                operation.view_scale[2] = +1.0f;

                // Save the rendering parameters.
                GLenum filtering = m_RenderingChain->filteringMethod ();

                // Setup the finest rendering.
                m_RenderingChain->setFilteringMethod (GL_LINEAR);

                // Render on a backbuffer.
                m_RenderingChain->render (
                    GL_COLOR_ATTACHMENT0_EXT,
                    &operation
                );

                // Restore the rendering parameters.
                m_RenderingChain->setFilteringMethod (filtering);
                target->unlock ();
            }
        } else {
            m_LastError = "FBO error.";
            logMessage (LOGIMPORTANCE_CRITICAL, m_LastError);
        }
    } else {
        m_LastError = "No rendering chain available.";
        logMessage (LOGIMPORTANCE_CRITICAL, m_LastError);
    }

    return target;
}

void Canvas::updateRenderingChain (void) {
    if (m_RenderingChain) {
        m_RenderingChain->clear ();

        // Insert all layers into a chain.
        logMessage (LOGIMPORTANCE_NEVERMIND, "Updating rendering chain...");
        foreach (ICanvasLayer * layer, m_Layers) {
            IRenderingChainLayer * rl = m_RenderingChain->addLayer (0.0f, 0.0f, 1.0f, 1.0f);
            if (rl) {
                rl->setOpacity  (layer->opacity ());
                rl->setLayerDpi (layer->dpi ());
                for (int i = 0; i < layer->size (); i++) {
                    rl->append (layer->at (i));
                }
            }
        }

        // Add a mixer layer if visible.
        if (m_MixerVisible && m_MixerLayer) {
            const float sx = (float)m_MixerWidth  / width ();
            const float sy = (float)m_MixerHeight / height ();
            IRenderingChainLayer * rl = m_RenderingChain->addLayer (0.1f * sx, 0.1f * sy, sx, sy);
            if (rl) {
                const float mixerbg[] = {
                    0.5f, 0.5f, 0.5f, 0.5f
                };

                rl->setOpacity          (m_MixerOpacity);
                rl->setBackgroundColor  (mixerbg);
                for (int i = 0; i < m_MixerLayer->size (); i++) {
                    rl->append (m_MixerLayer->at (i));
                }
            }
        }

        // Pass the bump map to the rendering chain.
        m_RenderingChain->setBumpMap (
                m_Papermap->bump (),
                m_Papermap->paperDpi ()
        );
    }
}

void Canvas::changesCommit (void) {
    m_Changed = FALSE;
    foreach (ICanvasLayer * layer, m_Layers) {
        layer->changesCommit ();
    }
}

bool Canvas::dry (void) {
    if (!(m_HasGLContext && m_PipelineDry))
        return FALSE;

    makeCurrent ();

    bool result = FALSE;

    // FIXME Current layer only?..
    // FIXME Must affect ONLY drying.
    foreach (ICanvasLayer * layer, m_Layers) {
        if (layer) {
            // Attach all channels to pipeline.
            for (int i = 0; i < layer->size (); i++) {
                IChannel * channel = layer->at (i);
                if (channel) {
                    channel->pokeObject ();
                    m_PipelineDry->pushChannel (channel);
                }
            }

            // And execute it.
            result |= m_PipelineDry->iterate (100.0f);
       }
    }

    repaint_constrained (rect ());
    return result;
}

/// Iterates through all channels of the
/// current layer, enumerates all patches,
/// and returns the patch with given index.
IArrayGridPatch * Canvas::patchIterator (int index) {
    if (m_CurrentLayer && index >= 0) {
        int index1 = index;
        for (int i = 0; i < m_CurrentLayer->size (); i++) {
            IChannel * channel = m_CurrentLayer->at (i);
            if (channel) {
                IArrayGrid * grid = channel->lock ();
                if (grid) {
                    const int sz = grid->nPatchesX () * grid->nPatchesY ();
                    if (index1 >= sz)
                        index1 -= sz;
                    else {
                        const int y = index1 / grid->nPatchesY ();
                        const int x = index1 % grid->nPatchesY ();
                        return grid->patch (x, y);
                    }
                }
            }
        }
    }

    return NULL;
}

void Canvas::selectionChanged (QRect selection) {
    m_rSelection = selection;
    repaint_constrained (rect ());
}

unsigned long Canvas::requiredBufferSize (void) const {
    const long canvasMetaSz = sizeof(qint32) * 2;

    long size = canvasMetaSz;
    foreach (ICanvasLayer * layer, m_Layers) {
        size += layer->requiredBufferSize ();
    }

    return size;
}

unsigned long Canvas::saveData (QDataStream * stream, unsigned long capacity) {
    m_LastError = NULL;
    if (capacity < requiredBufferSize ()) {
        m_LastError = g_err_BUFTOOSMALL;
        return 0;
    }

    unsigned long written = 0;

    // Write canvas metadata.
    stream->operator << ((qint32) m_Layers.size ());
    written += sizeof(qint32);

    // Write all canvas layers.
    foreach (ICanvasLayer * layer, m_Layers) {
        if (capacity > written)
            written += layer->saveData (stream, capacity - written);
        else {
            m_LastError = g_err_BUFTOOSMALL;
            logMessage (LOGIMPORTANCE_CRITICAL, "Cannot save the file because the capacity is not enough.");
        }
    }

    return written;
}

unsigned long Canvas::loadData (QDataStream * stream, unsigned long capacity) {
    m_LastError = NULL;
    if (capacity < (unsigned long)(requiredBufferSize () - sizeof(qint32) * 2)) {
        m_LastError = g_err_BUFTOOSMALL;
        return 0;
    }

    qint32 layers;
    unsigned long read = 0;

    // Read the canvas metadata.
    stream->operator >> (layers);
    read += sizeof(qint32);

    // Sanity check.
    if (layers > 128 || layers < 0)
        return 0L;

    // Read all layers.
    for (int i = 0; i < layers; i++) {
        // Read the layer metadata.
        qint32 width  = 0;
        qint32 height = 0;
        stream->operator >> (width);
        stream->operator >> (height);
        read += 2 * sizeof(qint32);

        // Sanity check.
        if (width  > 16 * 1024 || width  < 16 ||
            height > 16 * 1024 || height < 16)
            return 0L;

        // Create the layer.
        ICanvasLayer * layer = g_fac_SPAWN_CANVASLAYER (width, height, m_Helper);
        if (layer) {
            layer->setDpi (dpi ());

            // Read the layer data,
            if (capacity > read) {
                append (layer);
                read += layer->loadData (stream, capacity - read);
                layer->release ();
            }
        } else {
            return read;
        }
    }

    return read;
}

void Canvas::setMouseAction (int button, eDragMode action) {
    if (m_MouseHash.contains ((Qt::MouseButton)button)) {
        m_MouseHash.remove   ((Qt::MouseButton)button);
    }

    m_MouseHash.insert ((Qt::MouseButton)button, action);
}

eDragMode Canvas::getMouseAction (int button) {
    if (m_MouseHash.contains ((Qt::MouseButton)button)) {
        return m_MouseHash.value ((Qt::MouseButton)button);
    } else
        return DRAGSTATE_NOOP;
}

void Canvas::enableRenderingOptions (int flags) {
    if (m_RenderingChain) {
        m_RenderingChain->enableOptions (flags);
    }
}

void Canvas::disableRenderingOptions (int flags) {
    if (m_RenderingChain) {
        m_RenderingChain->disableOptions (flags);
    }
}
