/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#pragma once

//#ifdef QT_OPENGL_LIB
#   include "../../../glew/include/GL/glew.h"

#   include <QtGui>
#   include <QtOpenGL>

#include "commdefs.h"
#include "commtypes.h"
#include "commmath.h"
#include "refcounter.h"

#include "papermap.h"

#include "iCanvas.h"
#include "iInstrumentLockEx.h"
#include "iRenderingChain.h"
#include "preferences.h"


#include "scene.h"

#define MAX_CANVAS_LAYERS       64

enum eCanvasRq {
    RQ_RENDERINGCHAIN,
    RQ_PIPELINES,
    RQ_CEILING
};

/**
 * Define this to enable logging of
 * various information about the input
 * devices being used.
 */
#define DEBUG_INPUTDEVICE_INFORMATION

/**
 * Maximum number of canvas
 * updates per second.
 */
#define FPS_LIMIT           25


/**
 * Canvas widget represents the drawing area
 * that may contain several layers.
 */
class Canvas : public QGLWidget,
               public CRefCounter,
               public ICanvasDelegate,
               public ICanvas {
    Q_OBJECT
    _VERBOSEOBJECT
    _GLEW_ENABLED

    IRenderingChain   * m_RenderingChain;
    ICanvasLayer      * m_CurrentLayer;      /**< Current canvas layer.                  */
    IChannel          * m_CurrentChannel;    /**< Current canvas channel.                */
    IInstrumentProxy  * m_Instrument;        /**< Current instrument.                    */
    IConnector        * m_Connector;         /**< Interface to the server.               */
    IPaintingStyle    * m_PaintingStyle;     /**< Current painting style.                */
    ICanvasLayer      * m_MixerLayer;        /**< The separate layer for color mixing.   */
    IOpenGLHelper     * m_Helper;
    ISpline           * m_SplineFeedbackMu;
    IPipeline         * m_PipelineDry;       /**< Pipeline to perform the drying operation. */
    IImageTransaction * m_ActionControl;     /**< Undo-stack interface.                     */

    /**
     * Opened transaction associated with
     * the stroke.
     */
    ITransaction * m_CurrentTransaction;

    /**
     * Paper map descriptor.
     */
    Papermap * m_Papermap;
    Scene      m_Scene;

    /**
     * A map of requests sent
     * to avoid duplicates.
     */
    bool m_ReqMap[RQ_CEILING];

    int   m_Id;
    int   m_ImageDpi;
    int   m_PlaneWidth;
    int   m_PlaneHeight;
    int   m_MixerWidth;
    int   m_MixerHeight;
    float m_MixerOpacity;
    bool  m_HasGLContext;
    bool  m_Changed;
    bool  m_MixerVisible;
    bool  m_StrokeStarted;
    bool  m_TabletPresence;
    bool  m_CumulativeRectTrigger;
    float m_Gravity[2];

    bool  m_InStroke;
    bool  m_UseSmartStroke;
    float m_SmartStroke_mu;         /**< Current brush size multiplier.             */
    float m_LastReservoirSz;
    float m_StrokeDirection[2];     /**< Recorded stroke direction.                 */
    int   m_LineStrokeLength[2];

    float m_propMetaDataScale;      /**< Scaling factor for the metadata planes.    */
    float m_ViewScale[3];           /**< Image plane scaling.                       */
    float m_ViewTranslate[3];       /**< Image plane translation.                   */
    float m_ViewRotate[3];          /**< Image plane rotation.                      */
    float m_matModelView[3][3];     /**< Modelview matrix for image plane.          */
    float m_matModelView_inv[3][3]; /**< Inverse if a modelview matrix.             */

    /**
     * Pressure working range.
     * Should be obtained though calibration process.
     */
    float  m_PressureWRng[2];

    eDragMode m_DragMode;
    int       m_LastXPos;
    int       m_LastYPos;
    int       m_LastXPos_w;
    int       m_LastYPos_w;
    bool      m_LastPosInside_w;
    int       m_CurrentStrokeLength_x;
    int       m_CurrentStrokeLength_y;

    sTouch m_Stroke[1024];
    int    m_Stroke_len;
    int    m_Stroke_idx;

    /**
     * Current global status to render
     * an apropriate icon.
     */
    QHASGlobal::eStatus m_GlobalStatus;

    // Rendering parameters.

    sRect m_rViewport;
    sRect m_rTargetClip;
    QRect m_rCumulative;
    QRect m_rPointer;
    QRect m_rSelection;

    /**
     * List of all canvas layers.
     * Their order is important because
     * they'll be rendered in that sequence.
     */
    QList<ICanvasLayer *> m_Layers;
    QMutex  m_ReqMutex;
    QTimer  m_timerConsciousness;
    QTime   m_timerUpdate;

    /**
     * Actions table for mouse buttons.
     */
    QHash<Qt::MouseButton, eDragMode> m_MouseHash;

    /**
     * Last used pointer type for
     * the tablet device.
     */
    int m_LastPointer;

    /**
     * Id of the last used device.
     */
    qint64 m_LastDevUid;

    /**
     * List of devices that were
     * detected as active.
     */
    QList<qint64> m_Devices;


    Canvas (QWidget * parent, int width, int height, IOpenGLHelper * helper);
    virtual ~Canvas (void);

    /**
     * Locks the given instrument and applies
     * it to current data layer.
     * @return The invalidated rectangle.
     */
    QRect applyInstrument (IInstrumentProxy * instrument, ICanvasLayer * layer);
	QMutex instrumentMutex;

    /**
     * Re-compute to modelview matrix
     * and it's inverse according to
     * current transformation.
     */
    void rebuildMatrix (void);

    /**
     * Change the scrollers properties
     * according to current canvas plane
     * transformation.
     */
    void rebound (void);


    ISpline * rebuildSpline_fbmu (float rsz);

    /**
     * Request the canvas repainting.
     */
    void repaint_constrained (const QRect & rect);

    void finishStroke (void);

    /**
     * Render the canvas plane
     * onto the backbuffer.
     */
    void render (const QRect & rect);

	/**
	 * Set the interpolation step to
	 * R / STROKE_ISTEP_R, where R is
	 * a brush radius.
	 */
	int STROKE_ISTEP_R()
	{
		return 1 << PREF.strokeQuality;
	}


    virtual void suicide (void) {
        delete this;
    }

friend ICanvas * g_fac_SPAWN_CANVAS (QWidget * parent, int width, int height, IOpenGLHelper * helper);
friend ICanvas * g_fac_SPAWN_CANVAS (int width, int height, IOpenGLHelper * helper);
public:
    /******** ICANVAS IMPLANTS **********/

    virtual QWidget * asWidget (void) {
        return this;
    }

    /**
     * Get the real width of a
     * canvas plane.
     */
    virtual long realWidth (void) const {
        return m_PlaneWidth;
    }

    /**
     * Get the real height of a
     * canvas plane.
     */
    virtual long realHeight (void) const {
        return m_PlaneHeight;
    }

    /**
     * Change the canvas size.
     * If the canvas is not empty, it will try to
     * resize all the underlying data structures.
     */
    virtual void resize (long width, long height);

    /**
     * Erase all canvas contents.
     */
    virtual void erase (void);

    /**
     * Set the paper texture to canvas.
     */
    virtual void setPaperTexture (sPaper * texture);

    /**
     * Get current paper texture.
     */
    virtual const sPaper * paperTexture (void) const {
        return m_Papermap->paper ();
    }

    /**
     * Apply a given filter to all
     * data layers.
     */
    virtual void applyFilter (IFilter * filter);

    /**
     * Applies the given pipeline to all
     * data layers.
     * /return TRUE if the canvas has been changed.
     */
    virtual bool applyPipeline (IPipeline * pipe, float dt);

    /**
     * Returns the interface to Undo Stack object.
     */
    virtual IUndostackProxy * getUndoStack (void) {
        // FIXME Not implemented.
        return NULL;
    }

    /**
     * Set the gravity vector that
     * that will affect the water simulation.
     */
    virtual void setGravity (float x, float y) {
        m_Gravity[0] = x;
        m_Gravity[1] = y;
    }

    /**
     * Override the current painting style.
     * Pointer to interface must be always
     * available to the canvas.
     */
    virtual void setPaintingStyle (IPaintingStyle * style) {
        m_PaintingStyle = style;
    }

    /**
     * Get the current layer.
     * /return The pointer to layer interface or NULL.
     */
    virtual ICanvasLayer * currentLayer (void) {
        return m_CurrentLayer;
    }

    /**
     * Create a new layer with the dimentions of
     * canvas and return it without adding to the list.
     */
    virtual ICanvasLayer * allocateNewLayer (void);

    /**
     * Request the addition of a given channel
     * to the current layer if it is not already
     * present there.
     */
    virtual bool promoteChannel (FOURCC uid);

    /**
     * Show or hide the color mixer.
     */
    virtual void setMixerVisibility (bool visible);

    /**
     * Check if the stroke shaping is enabled.
     * This mode will try to dynamically scale the
     * brush if it is controlled my mouse to make the
     * stroke look more realistic.
     */
    virtual bool isSmartStrokeEnabled (void) const {
        return m_UseSmartStroke;
    }

    /**
     * Enable or disable the brush stroke shaping.
     */
    virtual void setSmartStrokeEnabled (bool enable) {
        m_UseSmartStroke = enable;
    }

    /**
     * Deallocate all objects and prepare the
     * canvas for termination.
     */
    virtual void deInit (void);

    /******** ISCROLLAREA IMPLANTS **********/

    virtual void setXTranslation (int x);
    virtual void setYTranslation (int y);
    virtual int addXTranslation (int x);
    virtual int addYTranslation (int y);
    virtual void getTranslation (int & x, int & y) const;

    virtual void setUniformScale (float scale);
    virtual float getUniformScale (void) const;
    virtual void multiplyScale (float scale);

    /**
     * Rotation angle must be given in degrees.
     */
    virtual void setRotation (float angle);
    virtual void addRotation (float angle);
    virtual float getRotation (void) const;

    /*********** QLIST WRAPPER **************/

    virtual void append   (ICanvasLayer * layer);
    virtual void prepend  (ICanvasLayer * layer);
    virtual void removeAt (int index);
    virtual void clear    (void);

    virtual int layersCount (void) {
        return m_Layers.size ();
    }

    virtual ICanvasLayer * at (int index);

    /**
     * Export the canvas contents as an image.
     * Returns an image that must be released by caller,
     * or NULL if any error ocurred.
     * The object is created with GL_RGB8 format and
     * GL_UNSIGNED_BYTE as a component type.
     */
    virtual IArray * asImage (void);

    /**
     * Export the canvas contents to a
     * specified target.
     */
    virtual IArray * asImage (IArray * target);

    /**
     * Set the mouse dragging action.
     */
    virtual void setDragMode (eDragMode mode) {
        m_DragMode = mode;
    }

    /**
     * Set the action for a mouse button.
     */
    virtual void setMouseAction (int button, eDragMode action);

    /**
     * Get the action for a mouse button.
     */
    virtual eDragMode getMouseAction (int button);

    /**
     * Set the image dpi.
     * This may also automatically change the dpi
     * for all layers.
     */
    virtual void setDpi (int dpi, bool layers);

    /**
     * Get the image dpi.
     */
    virtual int dpi (void) const {
        return m_ImageDpi;
    }

    /**
     * Get the current pressure working range.
     * @returns Pointer to min and max values.
     */
    virtual const float * pressureWorkingRange (void) const {
        return m_PressureWRng;
    }

    /**
     * Set the pressure working range.
     */
    virtual void setPressureWorkingRange (float minp, float maxp) {
        m_PressureWRng[0] = minp;
        m_PressureWRng[1] = maxp;
    }

    /**
     * Move all paint onto the dry layer.
     */
    virtual bool dry (void);

    /**
     * Enumerator for all patches of the
     * current layer.
     * @returns Patch pointer or NULL if index is out of range.
     */
    virtual IArrayGridPatch * patchIterator (int index);

    /**
     * Change the current instrument.
     */
    virtual void setInstrument (IInstrumentProxy * instrument);

    /**
     * Enable the rendering options for current
     * rendering chain.
     */
    virtual void enableRenderingOptions (int flags);

    /**
     * Disable the rendering options for current
     * rendering chain.
     */
    virtual void disableRenderingOptions (int flags);

    /****** ICANVASDELEGATE IMPLANTS ********/

    /**
     * Change the current selection.
     */
    virtual void setSelection (int x, int y, int w, int h) {
        selectionChanged (QRect (x, y, w, h));
    }

    /**
     * Broadcast a new color to be
     * set as current.
     */
    virtual void pickColor (IPigmentMixture * pig) {
        emit colorPicked (pig);
    }

    /**
     * Paint a brush touch on a specific channel.
     */
    virtual void touch (IInstrumentLock * operation);

    /***** IDATACONTAINTER IMPLANTS *******/

    /**
     * Check if the container's content
     * was changed since the last commit.
     */
    virtual bool changesCheck (void) const {
        return m_Changed;
    }

    /**
     * Commit the changes. This resets the
     * result returned by changesCheck ().
     */
    virtual void changesCommit (void);

    /**
     * The length of buffer required
     * to save the cantainer's data.
     */
    virtual unsigned long requiredBufferSize (void) const;

    /**
     * Save the container's data into
     * the external buffer.
     * @return Number of bytes actually written.
     */
    virtual unsigned long saveData (QDataStream * stream, unsigned long capacity);

    /**
     * Load the data from  the external
     * buffer into container.
     * @return Number of bytes actually read.
     */
    virtual unsigned long loadData (QDataStream * stream, unsigned long capacity);

    /****** ISPAWNEDOBJECT IMPLANTS *******/

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

    virtual const char * myClassName (void) {
        return "Client::Canvas";
    }

    /***** ICONNECTORCALLBACK IMPLANTS *****/

    /**
     * Pass the connector interface
     * to the object.
     */
    virtual void connect (IConnector * caller);

    /**
     * A callback, called by connector when
     * the requested data is recieved.
     */
    virtual void complitedRequest (int param, int length, void * data);

    /******** IVERBOSEOBJECT IMPLANTS *********/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_LastError;
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Enable of disable logging.
     */
    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

    /********** QGLWIDGET IMPLANTS ************/

    virtual QSize minimumSizeHint (void) const {
        return QSize (128, 128);
    }

 signals:
    /**
     * Emited each time pointer is moved.
     */
    void tracking (float x,    float y,
                   float dxdt, float dydt,
                   float pressure,
                   bool  inside);

    /**
    * Emitted when user touches the
    * canvas with his dirty hands.
    */
    void userStartsPainting (void);

   /**
    * Emitted when user removes
    * his dirty hands from a canvas.
    */
    void userStopsPainting (void);

    /**
     * Emited to show that the canvas has
     * the OpenGL context inited from now on.
     */
    void glcontextObtained (void);

    /**
     * Emited to show that the canvas
     * doesn't provide the OpenGL context
     * anymore.
     */
    void glcontextLost (void);

    /**
     * Emitted when user changes the
     * input device. Zero is passed
     * for mouse.
     */
    void tabletDeviceChanged (qint64 uid);

    /**
     * Emitted when user changes the
     * pen on his tablet.
     */
    void tabletPenChanged (QTabletEvent::PointerType pen);

    /**
     * Emitted when the stroke has
     * significantly changed its direction.
     */
    void strokeTurns (void);

    /**
     * Emitted when we have a mouse over canvas.
     */
    void gainedFocus (void);

    /**
     * Emitted when the instrument picks a
     * color from canvas.
     */
    void colorPicked (IPigmentMixture * pig);

    /**
     * Emitted when the canvas has gained
     * a rendering chain object and is ready
     * to recieve the options.
     */
    void renderingOptsRequest (void);

private slots:

    /**
     * Poked when it is time to look around.
     */
    void poke (void);

    void reset (void);

public slots:
    /**
     * Inform the canvas that the global
     * status has been changed.
     */
    void statusChanged (QHASGlobal::eStatus status);

    /**
     * Inform the canvas that the instrument
     * has been changed.
     */
    void instrumentChanged (IInstrumentProxy * instrument) {
        setInstrument (instrument);
    }

    /**
     * Inform the canvas that the instrument size
     * has been changed.
     */
    void instrumentSizeChanged (IInstrumentProxy * instrument,
                                int radius);

    /**
     * Inform the canvas that the layer
     * has been changed.
     */
    void layerChanged (ICanvasLayer * layer);

    /**
     * Inform the canvas that the
     * paper texture has been changed.
     */
    void paperChanged (sPaper * paper) {
        setPaperTexture (paper);
    }

    void channelPromoted (FOURCC uid) {
        promoteChannel (uid);
    }

    void showMixer (void) {
        setMixerVisibility (TRUE);
    }

    void hideMixer (void) {
        setMixerVisibility (FALSE);
    }

    /**
     * Pass the connector interface
     * to the object.
     */
    void connected (IConnector * caller) {
        connect (caller);
    }

    void disconnected (void) {
        SAFE_RELEASE (m_Connector);
    }

    /**
     * Set the new selection.
     * Empty rect clears the selection.
     */
    void selectionChanged (QRect selection);

protected:
    void initializeGL (void);
  //void paintGL      (void);
    void paintEvent   (QPaintEvent * event);
    void resizeGL     (int width, int height);

    void mousePressEvent   (QMouseEvent  * event);
    void mouseReleaseEvent (QMouseEvent  * event);
    void mouseMoveEvent    (QMouseEvent  * event);
    void enterEvent        (QEvent       * event);
    void leaveEvent        (QEvent       * event);
    void tabletEvent       (QTabletEvent * event);

    /**
     * Rebuild the rendering chain when
     * the channel list has chaned.
     */
    void updateRenderingChain (void);

    /**
     * Process the instrument movement event.
     */
    void instrumentMoves (int x0, int y0, bool pressed);
};

//#endif
