/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CANVASOPERATION_H
#define CANVASOPERATION_H


/**
 * Extended operation descriptor.
 */
class CanvasOperation : public IInstrumentLockEx {
    IInstrumentProxy  * m_Instrument;
    IInstrumentLock   * m_Source;
    IArray            * m_Paper;
    IPaintingStyle    * m_PaintingStyle;
    ITransaction      * m_Transaction;

    float m_Velocity[2];
    float m_FeedbackMu;
    int   m_PaperDpi;

friend class Canvas;
public:

    /******* IINSTRUMENTLOCK IMPLANTS ***********/

    /**
     * Drawing privitive color texture.
     */
    virtual IArray * primitiveColor (void) {
        return m_Source->primitiveColor ();
    }

    /**
     * Drawing privitive shape texture.
     */
    virtual IArray * primitiveShape (void) {
        return m_Source->primitiveShape ();
    }

    /**
     * Drawing privitive geometry
     * as a vertex buffer.
     */
    virtual GLuint primitiveGeometry (int & nVertices) const {
        return m_Source->primitiveGeometry (nVertices);
    }

    /**
     * Rotation angle of the primitive
     * around the Z axis.
     */
    virtual int primitiveRotation (void) const {
        return m_Source->primitiveRotation ();
    }

    /**
     * Returns the components mask that contain
     * zeros for unused components and the proper
     * multipliers for the rest.
     */
    virtual const float * flowControlComponentMask (void) const {
        return m_Source->flowControlComponentMask ();
    }

    /**
     * Instrument bleeding rate.
     */
    virtual float bleedingRate (void) const;

    /**
     * Instrument sucking rate.
     */
    virtual float suckingRate (void) const;

    /**
     * Instrument reservoir size determines the
     * threshold delaying the color feedback.
     */
    virtual float reservoirSize (void) const {
        return m_Source->reservoirSize ();
    }

    /**
     * Get the texture sensitivity of
     * the instrument in a range [0, 1].
     */
    virtual float stipplingSensitivity (void) const;

    /**
     * Perlin noise parameters block.
     */
    virtual const sNoiseParameters * noiseParameters (void) const {
        return m_Source->noiseParameters ();
    }

    /**
     * Get the scaling factor for
     * the instrument.
     */
    virtual float scalingFactor (void) const {
        return m_Instrument->scaleModifier () *
               m_Source->scalingFactor ();
    }

    /**
     * OR-ed set of flags specifying
     * what data planes should be affected by instrument.
     */
    virtual int affectsDataPlanes (void) const {
        return m_Source->affectsDataPlanes ();
    }

    virtual void release (void) {
    }

    /****** IINSTRUMENTLOCKEX IMPLANTS **********/

    virtual IArray * paperTexture (void) {
        return m_Paper;
    }

    /**
     * Dpi value for the paper texture.
     */
    virtual int paperDpi (void) const {
        return m_PaperDpi;
    }

    /**
     * Returns the pointer to painting
     * style description interface.
     */
    virtual IPaintingStyle * paintingStyle (void) {
        return m_PaintingStyle;
    }

    /**
     * Current instrument velocity.
     */
    virtual const float * velocity (void) const {
        return m_Velocity;
    }

    /**
     * Interface to current transaction if available.
     */
    virtual ITransaction * transaction (void) {
        return m_Transaction;
    }
};

#endif // CANVASOPERATION_H
