/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef BRUSHLIST_H
#define BRUSHLIST_H

#include "commdefs.h"
#include "commtypes.h"
#include "commmath.h"
#include "refcounter.h"

#include <QtGui>
#include <QPainter>
#include <QMetaType>
#include <QItemDelegate>
#include <QListWidget>
#include <QHash>

//#include "../hgwidgets/Rotary.h"

//#ifdef QT_OPENGL_LIB
#   include "../../../glew/include/GL/glew.h"

#   include "iInstrument.h"
#   include "iInstrumentProxy.h"
#   include "iInstrumentList.h"
#   include "iCanvasLayer.h"
#   include "ui_brushlist.h"
#   include "qhaswidgets.h"
#   include "instrumentconfig.h"

/**
 * An instrument data for
 * QListView aggregation.
 */
class InstrumentPreview {
public:
    InstrumentPreview (void) {
        m_Instrument = NULL;
    }

    InstrumentPreview (IInstrumentProxy * instrument) {
        m_Instrument = instrument;
    }

    ~InstrumentPreview (void) {
    }

    /**
     * Returns the generated instrument image.
     */
    const QImage * image (void) const {
        return reinterpret_cast<const QImage*>(m_Instrument->qimage ());
    }

    /**
     * Return the pointer to associated
     * instrument interface.
     */
    IInstrumentProxy *instrument (void) const {
        return m_Instrument;
    }

private:
    IInstrumentProxy * m_Instrument;
};

Q_DECLARE_METATYPE (InstrumentPreview);


/**
 * A delegate class to paint the
 * instrument items in the list.
 */
class InstrumentDelegate : public QItemDelegate {
    Q_OBJECT

public:
    InstrumentDelegate (void) {
    }

    /**
     * Draw the instrument preview
     * using pre-rendered data.
     */
    void paint (QPainter * painter, const QStyleOptionViewItem & option,
        const QModelIndex & index) const;

    /**
     * Returns item size.
     */
    QSize sizeHint (void) const {
        return QSize (256, 32);
    }

    /**
     * Returns item size.
     */
    QSize sizeHint (const QStyleOptionViewItem &option,
                    const QModelIndex &index ) const {
        Q_UNUSED (option);
        Q_UNUSED (index);
        return sizeHint ();
    }

    void setModelData (QWidget * editor, QAbstractItemModel * model,
        const QModelIndex & index) const {
        Q_UNUSED (editor);
        Q_UNUSED (model);
        Q_UNUSED (index);
        return;
    }
};

/**
 * Instrument list.
 */
class Brushlist : public QFrame,
                  public IInstrumentList,
                  public CRefCounter {
    Q_OBJECT;
    Q_DISABLE_COPY (Brushlist);

public:
    explicit Brushlist (QWidget *parent = NULL);
    virtual ~Brushlist (void);

    /********* IINSTRUMENTLIST IMPLANTS ***********/

    virtual QWidget * asWidget (void) {
        return this;
    }

    /**
     * Add an instrument to
     * the end of the list.
     */
    virtual void append (IInstrument * instrument);

    /**
     * Add an instrument to
     * the beginning of the list.
     */
    virtual void prepend (IInstrument * instrument);

    /**
     * Remove an instrument at
     * given position.
     */
    virtual void removeAt (int index);

    /**
     * Remove the specified instrument
     * from list.
     */
    virtual void remove (IInstrument * target);

    /**
     * Returns the instrument at given position
     * or NULL if position is illegal ar empty.
     */
    virtual IInstrument * at (int index) {
        if (index >= 0 && index < m_Handles.size ()) {
            return m_Handles.at (index);
        } else {
            return NULL;
        }
    }

    /**
     * Removes all instruments from list.
     */
    virtual void clear (void);

    /**
     * Returns the currently
     * selected instrument.
     */
    virtual IInstrumentProxy * current (void) {
        return m_Current;
    }

    /**
     * Promote all channels needed by
     * the current instrument.
     */
    virtual void promoteChannels (void);

    /**
     * Change the brush scaling factor and
     * update the corresponding widget.
     */
    virtual void chageBrushScale (float d);

    /***** ICONNECTORCALLBACK IMPLANTS *****/

    /**
     * Pass the connector interface
     * to the object.
     */
    virtual void connect (IConnector * caller);

    /**
     * A callback, called by connector when
     * the requested data is recieved.
     */
    virtual void complitedRequest (int param, int length, void * data);

    /******* ISPAWNEDOBJECT IMPLANTS *******/

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Client::InstrumentList";
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
        if (m_Current) {
            promoteChannels ();
            emit instrumentChanged (m_Current);
        }
    }

    /******** IVERBOSEOBJECT IMPLANTS *********/

    /**
     * Returns the description of last error
     * or NULL if no error occured.
     */
    virtual const char * lastError (void) const {
        return m_LastError.toUtf8 ().data ();
    }

    /**
     * Attach the logger to an object.
     */
    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    /**
     * Enable of disable logging.
     */
    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    /**
     * Check if the logging is enabled.
     */
    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

public slots:
    /**
     * Notify the list when the
     * instrument pressure is changed.
     */
    void watchPressure (float pressure) {
        m_LastPressure = pressure;
    }

    /**
     * Notify the list when the
     * canvas layer is changed to
     * promote the instrument channels.
     */
    void canvasLayerChanged (ICanvasLayer * layer);

    /**
     * Notify the list when the user
     * starts drawing anything on a canvas.
     */
    void userStartsDrawing (void);

    /**
     * Notify the list when the user
     * stop drawing on a canvas.
     */
    void userStopsDrawing (void);

    /**
     * Notify the list when the
     * pen has been changed.
     */
    void tabletPenChanged (QTabletEvent::PointerType pen);

    /**
     * Notify the widget when user changes
     * the input device. Zero is passed
     * for mouse.
     */
    void tabletDeviceChanged (qint64 uid);

    /**
     * Show or hide the instrument
     * configuration panel.
     */
    void showConfigPanel (bool on);

signals:
    /**
     * Emitted when user selects
     * a different instrument.
     */
    void instrumentChanged (IInstrumentProxy * instrument);

    /**
     * Emitted when user changes
     * the instrument size.
     */
    void instrumentSizeChanged (IInstrumentProxy * instrument,
                                int radius);

    /**
     * Emitted to promote a channel
     * requested by an instrument.
     */
    void channelPromoted (FOURCC channel_uid);

private slots:
    void instrumentListSelectionChanged (int row);

protected:
    virtual void changeEvent (QEvent *e);

    /**
     * A call to logger object.
     */
    void logMessage (eLOG_IMPORTANCE importance, const char * msg, ...) {
        if (msg && m_Logger && m_LoggingEnabled) {
            va_list val;
            va_start (val, msg);
            m_Logger->write (importance, this, msg, val);

            if (importance == LOGIMPORTANCE_ERROR ||
                importance == LOGIMPORTANCE_CRITICAL)
                m_LastError = msg;
        }
    }

    virtual void suicide (void) {
        delete this;
    }

private:
    Ui::Brushlist    * m_ui;
    IInstrumentProxy * m_Current;
    ILogger          * m_Logger;
    IConnector       * m_Connector;

    /**
     * Popup panel to configure the instrument.
     */
    InstrumentConfig   m_ConfigPanel;

    QList<IInstrumentProxy*>    m_Handles;
    QHash<int, int>             m_StatesHash;

    int   m_Id;
    int   m_StylusId;
    float m_LastPressure;

    QString m_LastError;
    BOOL    m_LoggingEnabled;
};

//#endif // QT_OPENGL_LIB
#endif // BRUSHLIST_H

