/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "commtypes.h"
#include "commdefs.h"
#include "refcounter.h"

#include "worker.h"

#include "iOpenGLHelper.h"
#include "iCanvas.h"
#include "iConnector.h"
#include "iInstrumentList.h"
#include "iLayersList.h"
#include "iPalette.h"
#include "iWorkspace.h"
#include "iSkin.h"
#include "qhaswidgets.h"
#include "loggersignal.h"
#include "mymdisubwindow.h"

#include <QtGui>
#include <QVBoxLayout>
#include <QGLWidget>
#include <QMdiArea>
#include <QFile>
#include <QHash>
#include <QTimer>

#define STATUSBAR_MESSAGE_TIMEOUT           4096


namespace Ui {
    class MainWindowClass;
}

struct sBgTask {
    QString name;
    bool    blocker;
};

class MainWindow : public QMainWindow {
    Q_OBJECT

public:
    MainWindow (QWidget *parent = 0, Logger_signal * logger = NULL);
   ~MainWindow (void);

   bool isSimulationStarted (void) const;

public slots:
   void canvasTrack (float x,    float y,
                     float dxdt, float dydt,
                     float pressure,
                     bool  inside);

   void connected       (void);
   void disconnected    (void);
   void closeEvent      (QCloseEvent *event);
   void glContextGained (void);
   void glContextLost   (void);
   void colorChanged    (IPigmentMixture * pig, float amount, eColorMetod source = COLORMETHOD_KMSAMPLE);
   void colorChanged    (IPigmentMixture * pig);
   void colorChangedByColorPicker (IPigmentMixture * pig, float amount, eColorMetod source = COLORMETHOD_KMSAMPLE);
   void preferences();
   void backgroundTask (QString name, int percent, bool blocker);

private slots:
   void startSimulation (void);
   void stopSimulation  (void);
   void open            (void);
   void create          (void) {newFile ();}
   void clean           (void);
   bool save            (void);
   bool saveAs          (void);
   void about           (void);
   void logmessage      (eLOG_IMPORTANCE importance, const char * str);
   void reloadShaders   (void);
   void notAssigned     (void);
   void dockVisibility  (bool visible);
   void clearCanvas     (void);
   void dryCanvas       (void);
   void paintingStyleChanged(void);
   void changeCanvasScale (double scale);

   void userStartedPainting    (void);
   void userStoppedPainting    (void);
   void instrumentChanged      (IInstrumentProxy * instrument);
   void toolInstrumentChanged  (IInstrumentProxy * instrument);
   void strokeDirectionChanged (void);

   /**
    * Update various stupid controls
    * by timer.
    */
   void consciousness (void);

   void closeSubWindow(bool *);

   /**
     * Indicate that the work status
     * has been changed.
     */
   void statusChanged   (QHASGlobal::eStatus status);
   void resetDragStates (void);
   void canvasPreview   (QImage * image);

   void renderingOptsChanged   (void);
   void renderingOptsRequested (void);

protected:
   virtual void wheelEvent      (QWheelEvent * event);
   virtual void keyPressEvent   (QKeyEvent * event);
   virtual void keyReleaseEvent (QKeyEvent * event);

private:
    Ui::MainWindowClass * ui;
    const QString m_APP_TITLE;

    /**
     * Worker object that
     * performs some background tasks.
     */
    Worker * m_Worker;

    /**
     * Skin object.
     */
    ISkin * m_Skin;

    IWork           * m_wrkCanvasIterate;
    IWork           * m_wrkCanvasPreview;
    IConnector      * m_Connector;
    IOpenGLHelper   * m_Helper;
    IWorkspace      * m_Workspace;      /**< Workspace object.          */
    ICanvas         * m_Canvas;         /**< Canvas widget.             */
  //ICanvas         * m_Mixer;          /**< Color mixer widget.        */
    IInstrumentList * m_InstrumentList; /**< Instrument list widget.    */
    ILayersList     * m_LayersList;     /**< Layers list widget.        */
    IPaperSwitch    * m_PaperSwitch;    /**< Paper chooser widget.      */
    IPalette        * m_Palette;        /**< Palette widget.            */
    IPalette        * m_dynPalette;     /**< Dynamic Palette widget.    */
    ILogger         * m_Logger;

    bool m_RestoreInstrumentColor;
    bool m_Tracked;
    bool m_HasGLContext;
    bool m_UpdateBeep;
    bool m_Block;
    int  m_LastTrackX;
    int  m_LastTrackY;

    QString       m_curFile;
    QVBoxLayout * m_layoutCanvas;
    QMdiArea    * m_MdiArea;
    QToolBar    * m_FileToolBar;
    QToolBar    * m_EditToolBar;
    QToolBar    * m_WaterToolBar;
    QShortcut   * m_Shortcut_ClearCanvas;
    QTimer        m_TimerConsciousness;

    /**
     * List of background task names
     * that should be indicated by the window.
     */
    QList<sBgTask *> m_ThreadTasks;

    MyMdiSubWindow * m_MdiSub;
    WheelAction m_WheelCurrentModifier;

    /**
     * Table of modifier-keys and their actions.
     */
    QHash<int, eDragMode> m_ModifierKeys;
    QHash<int, WheelAction> m_ModifierKeysWheel;

    /**
     * Table of mouse-buttons and their actions.
     * NOT USED YET.
     */
    QHash<int, eDragMode> m_MouseKeys;

    void createToolbars  (void);
    void createCanvas    (void);
    void createShortcuts (void);
    void createHotkeys   (void);
    void updateTitle     (void);
    bool maybeSave       (void);
    bool saveFile        (const QString &fileName);
    bool loadFile        (const QString &fileName);
    void setCurrentFile  (const QString &fileName);
    void newFile         (void);
    void closeWorkspace  (void);
    void bindWorkspace   (IWorkspace * workspace);
  //void setStatusTip    (const QString & tip);
  //QString statusTip    (void) const;

    void readSettings    (void);
    void writeSettings   (void);

    bool colorChangedByMainWindow;

    // replace for modifiers keys by modifiers flags
    void checkModifiers(int & key);

#ifdef GLEW_MX
    GLEWContext * glewGetContext (void) {
         return m_Helper->context ();
     }
#endif
};

#endif // MAINWINDOW_H
