/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "mymdisubwindow.h"
#include "preferences.h"

bool iarrayToFileRgba (IArray * i, QString path);


void MainWindow::closeWorkspace (void) {
    // Stop and release all tasks.
    SAFE_RELEASE (m_wrkCanvasIterate);
    SAFE_RELEASE (m_wrkCanvasPreview);
    if (m_Worker)
        m_Worker->terminate ();

    if (m_Canvas) {
        // Detach all subwindows.
        if (m_MdiSub) {
            m_MdiSub->setWidget (NULL);
            m_MdiArea->removeSubWindow (m_MdiSub);
            SAFE_DELETE (m_MdiSub);
        }

        m_Canvas->asWidget ()->setVisible (FALSE);
        m_Canvas->clear ();
    }

    // Release the workspace.
    if (m_Workspace) {
        SAFE_RELEASE (m_Workspace);
    }

    // Detach the painting style descriptor.
    // FIXME Use QActionGroup ?..
    ui->colorPicker->setPaintingStyle   (NULL);
    ui->paintingStyle->setPaintingStyle (NULL);
    m_Worker->setPaintingStyle          (NULL);

    // Disable the workspace-related events.
    ui->actionSave->setEnabled     (FALSE);
    ui->actionSave_as->setEnabled  (FALSE);
    ui->actionClose->setEnabled    (FALSE);
}

void MainWindow::bindWorkspace (IWorkspace * workspace) {
    m_Workspace = workspace;
    if (m_Workspace && m_Canvas) {
        m_Canvas->setSmartStrokeEnabled (PREF.smartStroke);

        // Add the background layer to canvas
        // if it's empty.
        if (m_Canvas->layersCount () == 0) {
            ICanvasLayer * layer = m_Canvas->allocateNewLayer ();
            m_Canvas->append (layer);
            layer->setName (tr ("Background"));
            layer->release ();
        }

        // Attach the canvas to worker object:
        // * Iterate.
        m_wrkCanvasIterate = m_Worker->addCanvasUpdate (
            m_Connector,
            m_Canvas,
            m_Helper
        );
        if (m_wrkCanvasIterate) {
            m_wrkCanvasIterate->addRef ();
        }

        // Attach the canvas to worker object:
        // * Preview.
        m_wrkCanvasPreview = m_Worker->addLayerPreview (
            m_Connector,
            m_Canvas,
            m_Canvas->currentLayer(),
            m_Helper
        );
        if (m_wrkCanvasPreview) {
            m_wrkCanvasPreview->addRef ();
            connect (dynamic_cast<QObject*>(m_wrkCanvasPreview),
                     SIGNAL(previewUpdated (QImage*)),
                     this,
                     SLOT(canvasPreview(QImage*))
            );
        }

        // Share the painting style descriptor.
        ui->colorPicker->setPaintingStyle   (m_Workspace->paintingStyle ());
        ui->paintingStyle->setPaintingStyle (m_Workspace->paintingStyle ());
        m_Worker->setPaintingStyle          (m_Workspace->paintingStyle ());
        m_Workspace->attachPaperSwitch      (m_PaperSwitch);

        // Sync the painting style widget.
        ui->paintingStyle->sync (m_Workspace->paintingStyle ()->closestPreset ());

        m_PaperSwitch->pokeObject ();
        m_InstrumentList->pokeObject ();

        // Add a canvas to MDI area..
        if (m_MdiArea) {
            m_MdiSub = new MyMdiSubWindow(m_MdiArea, Qt::SubWindow|Qt::CustomizeWindowHint|Qt::WindowMaximizeButtonHint);//|Qt::WindowTitleHint|Qt::WindowMinMaxButtonsHint);
            m_MdiSub->setWidget    (m_Canvas->asWidget ());
            m_MdiSub->setAttribute (Qt::WA_DeleteOnClose);
            m_MdiSub->showMaximized();
            connect (m_MdiSub, SIGNAL(allowClosing(bool*)), this, SLOT(closeSubWindow(bool*)));

            m_Canvas->asWidget ()->setVisible (TRUE);
        }

        m_Worker->start ();

        ui->actionSave->setEnabled     (TRUE);
        ui->actionSave_as->setEnabled  (TRUE);
        ui->actionClose->setEnabled    (TRUE);
    }
}

bool MainWindow::saveFile (const QString &fileName) {
    if (!m_Workspace)
        return FALSE;

    bool result = FALSE;

    // Export as image.
    if (fileName.endsWith (".jpg",  Qt::CaseInsensitive) ||
        fileName.endsWith (".jpeg", Qt::CaseInsensitive) ||
        fileName.endsWith (".png",  Qt::CaseInsensitive) ||
        fileName.endsWith (".bmp",  Qt::CaseInsensitive)) {
        IArray * i = m_Canvas->asImage ();
        if (i == NULL) {
            QMessageBox::warning (
                this,
                tr ("QAquarelle. saving error."),
                m_Canvas->lastError ()
            );
        } else {
            result = iarrayToFileRgba (i, fileName);

            SAFE_RELEASE (i);
            if (result) {
                statusBar()->showMessage (tr ("Image saved."), STATUSBAR_MESSAGE_TIMEOUT);
            } else {
                statusBar()->showMessage (tr ("Failed to save the image."), STATUSBAR_MESSAGE_TIMEOUT);
            }

            return result;
        }
    } else
    // Save a binary data.
    if (fileName.endsWith (".qaq",  Qt::CaseInsensitive)) {
        long written = 0;
        QFile f (fileName);
        if (f.open (QFile::WriteOnly)) {
            if (f.isWritable ()) {
                QDataStream stream (&f);
                written += m_Workspace->saveData (&stream, (DWORD)-1);
            }
        }

        if (written < 1) {
            if (!f.isWritable ()) {
                QMessageBox::warning (
                    this,
                    tr ("QAquarelle saving error."),
                    tr ("I'm sorry, but I cannot open the file for writing: ") + fileName
                );
            } else {
                QMessageBox::warning (
                    this,
                    tr ("QAquarelle saving error."),
                    tr ("I'm sorry, but I couldn't save the workspace because of some error... =(")
                );
            }

            f.close ();
        }
        else {
            f.close ();
            statusBar()->showMessage (tr ("File saved (%1 MBytes).").arg (
                                      (float)written / 1024.0f / 1024.0f),
                                      STATUSBAR_MESSAGE_TIMEOUT);
            return TRUE;
        }
    } else
        QMessageBox::warning (
            this,
            tr ("QAquarelle saving error."),
            tr ("You have specified the file type that is unknown to me... I'm sorry.")
        );

    return FALSE;
}

bool MainWindow::loadFile (const QString &fileName) {
    // Create an empty workspace.
    IWorkspace * workspace = g_fac_SPAWN_WORKSPACE (m_Canvas, m_Helper);
    if (workspace) {
        workspace->attachLogger      (m_Logger);
        workspace->enableLogging     (TRUE);
        workspace->attachPaperSwitch (m_PaperSwitch);

        int read = 0;
        QFile f (fileName);
        if (f.open (QFile::ReadOnly)) {
            if (f.isReadable ()) {
                QDataStream stream (&f);
                read += workspace->loadData (&stream, (DWORD)-1);
            }
        }

        f.close ();

        if (read) {
            bindWorkspace (workspace);
            statusBar()->showMessage (tr ("File loaded (%1 MBytes).").arg (
                                      (float)read / 1024.0f / 1024.0f),
                                      STATUSBAR_MESSAGE_TIMEOUT);
            return TRUE;
        } else {
            SAFE_RELEASE (workspace);
        }
    }

    return FALSE;
}

