/***************************************************************************
 *   Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>              *
 *                                                                         *
 *   This file is a part of QAquarelle project.                            *
 *                                                                         *
 *   QAquarelle is free software; you can redistribute it and/or modify    *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "workspace.h"
#include "iPigmentDescription.h"
#include <math.h>

static const char * g_err_NOTHINGTOSAVE  = "There is nothing to save.";
static const char * g_err_INVALIDVERSION = "File version mismatch (deprecated format).";
static const char * g_err_INVALIDFILE    = "File has a wrong format.";


IWorkspace * g_fac_SPAWN_WORKSPACE (ICanvas * canvas, IOpenGLHelper * helper,
                                    long width, long height, int dpi) {
    return new Workspace (canvas, helper, width, height, dpi);
}

Workspace::Workspace (ICanvas * canvas, IOpenGLHelper * helper,
                      long width, long height, int dpi = 100) {
    m_Id = ISpawnedObject::allocateId ();

    _register (this);
    m_Width  = width;
    m_Height = height;
    m_Helper = helper;
    m_Canvas = canvas;
    if (m_Canvas) {
        m_Canvas->addRef ();
    }

    m_PaperMonster   = NULL;
    m_Logger         = NULL;
    m_LastError      = NULL;
    m_LoggingEnabled = FALSE;
    m_IOVersionMajor = 0;
    m_IOVersionMinor = 3;

    m_WorkspaceChanged = TRUE;
    m_WorkspaceCreated = FALSE;

    if (helper)
        helper->addRef ();

    // Init the painting style.
    fromPreset (PAINTINGSTYLE_WATERCOLOURS);

    if (width && height) {
        create (width, height, dpi);
    }
}

Workspace::~Workspace (void) {
    _unregister (this);

    SAFE_RELEASE (m_Canvas);
    SAFE_RELEASE (m_Helper);
}

void Workspace::create (long width, long height, int dpi) {
    m_LastError = NULL;

    if (!m_WorkspaceCreated) {
        m_Width  = width;
        m_Height = height;

        // Reinit the canvas.
        if (m_Canvas) {
            m_Canvas->clear ();
            m_Canvas->resize (width, height);
            m_Canvas->setPaintingStyle (paintingStyle ());
            m_Canvas->setDpi (dpi);
        }

        m_WorkspaceCreated = TRUE;
    }
}

/// Init the painting style from preset.
void Workspace::fromPreset (ePAINTINGSTYLE_PRESET preset) {
    m_LastError = NULL;

    m_ps_Opacity         = IPigmentDescription::opacity     (preset);
    m_ps_Viscosity       = IPigmentDescription::viscosity   (preset);
    m_ps_DryingSpeed     = IPigmentDescription::dryingspeed (preset);
}

unsigned long Workspace::requiredBufferSize (void) const {
    unsigned long sz = 5 * sizeof(qint32);   // Metadata.
    if (m_Canvas) {                          // Canvas data.
        sz += m_Canvas->requiredBufferSize ();
    }

    return sz;
}

unsigned long Workspace::saveData (QDataStream * stream, unsigned long capacity) {
    m_LastError = NULL;
    if (capacity < requiredBufferSize ()) {
        return 0L;
    }

    unsigned long written = 0L;

    // Write the magic.
    stream->operator << ((qint32) g_MAKE_FOURCC ('Q', 'A', 'C', 'K'));
    stream->operator << ((qint32) m_IOVersionMajor);
    stream->operator << ((qint32) m_IOVersionMinor);
    written += sizeof(qint32) * 3;

    // Write the painting style information.
    stream->operator << (m_ps_DryingSpeed);
    stream->operator << (m_ps_Opacity);
    stream->operator << (m_ps_Staining);
    stream->operator << (m_ps_Viscosity);
    written += sizeof(float) * 4;

    if (m_Canvas) {
        if (capacity > written) {
            stream->operator << ((qint32) m_Canvas->realWidth  ());
            stream->operator << ((qint32) m_Canvas->realHeight ());
            stream->operator << ((qint32) m_Canvas->dpi ());
            written += sizeof(qint32) * 3;
            written += m_Canvas->saveData (stream, capacity - written);
        }
    } else {
        stream->operator << ((qint32) 0);
        stream->operator << ((qint32) 0);
        stream->operator << ((qint32) 0);
        written += sizeof(qint32) * 3;

        m_LastError = g_err_NOTHINGTOSAVE;
        logMessage (LOGIMPORTANCE_WARNING, m_LastError);
    }

    if (m_PaperMonster)
        written += m_PaperMonster->saveData (stream, capacity - written);

    return written;
}

unsigned long Workspace::loadData (QDataStream * stream, unsigned long capacity) {
    m_LastError = NULL;
    if (capacity < requiredBufferSize ()) {
        return 0L;
    }

    unsigned long read = 0L;

    quint32 magic;
    qint32 v0;
    qint32 v1;

    // Read and validate the metadata.
    stream->operator >> (magic);
    stream->operator >> (v0);
    stream->operator >> (v1);
    read += sizeof(qint32) * 3;

    if (magic != g_MAKE_FOURCC ('Q', 'A', 'C', 'K')) {
        m_LastError = g_err_INVALIDFILE;
        logMessage (LOGIMPORTANCE_WARNING, m_LastError);
        return 0L;
    }

    if (v0 != m_IOVersionMajor ||
        v1 != m_IOVersionMinor) {
        m_LastError = g_err_INVALIDVERSION;
        logMessage (LOGIMPORTANCE_WARNING, m_LastError);
        return 0L;
    }

    // Read the painting style information.
    stream->operator >> (m_ps_DryingSpeed);
    stream->operator >> (m_ps_Opacity);
    stream->operator >> (m_ps_Staining);
    stream->operator >> (m_ps_Viscosity);
    read += sizeof(float) * 4;

    // Sanity check.
    if (m_ps_DryingSpeed < 0.0f || m_ps_DryingSpeed > 1.0f ||
        m_ps_Opacity     < 0.0f || m_ps_Opacity     > 1.0f ||
        m_ps_Staining    < 0.0f || m_ps_Staining    > 1.0f ||
        m_ps_Viscosity   < 0.0f || m_ps_Viscosity   > 1.0f)
        return 0L;

    qint32 width;
    qint32 height;
    qint32 dpi;
    stream->operator >> (width);
    stream->operator >> (height);
    stream->operator >> (dpi);
    read += sizeof(qint32) * 3;

    // Sanity check.
    if (dpi > 4096 || dpi < 1)
        return 0L;

    create (width, height, dpi);
    if (m_Canvas && capacity > read)
        read += m_Canvas->loadData (stream, capacity - read);

    if (m_PaperMonster)
        read += m_PaperMonster->loadData (stream, capacity - read);

    return read;
}

ePAINTINGSTYLE_PRESET Workspace::closestPreset (void) const {
    for (int i = 0; i < PAINTINGSTYLES_COUNT; i++) {
        const ePAINTINGSTYLE_PRESET preset = (ePAINTINGSTYLE_PRESET) i;

        if (IS_SIGNIFICANT (fabsf (m_ps_DryingSpeed - IPigmentDescription::dryingspeed (preset))))
            continue;
        if (IS_SIGNIFICANT (fabsf (m_ps_Opacity - IPigmentDescription::opacity (preset))))
            continue;
        if (IS_SIGNIFICANT (fabsf (m_ps_Staining - IPigmentDescription::stainingPower (preset))))
            continue;
        if (IS_SIGNIFICANT (fabsf (m_ps_Viscosity - IPigmentDescription::viscosity (preset))))
            continue;

        return preset;
    }

    return PAINTINGSTYLE_CUSTOM;
}
