/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "arraygrid.h"
#include "arraygrid_patch.h"
#ifdef DEBUG_COLORED_PATCHES
#   include <stdlib.h>
#endif


IArrayGrid * g_fac_SPAWN_ARRAYGRID (IOpenGLHelper * helper, LONG width, LONG height,
                                    GLenum format, GLenum type) {
    return new ArrayGrid (helper, width, height, format, type);
}

ArrayGrid::ArrayGrid (IOpenGLHelper * helper, long width, long height, GLenum format, GLenum type) {
    PROFILE_FUNC();

    m_PlaneWidth  = width;
    m_PlaneHeight = height;
    m_ElementSz   = 0;

    m_Id = ISpawnedObject::allocateId ();
    ISpawnedObject::_register (this);

    m_Helper = helper;
    if (m_Helper) {
        m_Helper->addRef ();

        QMutexLocker (m_Helper->mutex ());
        m_PatchSize = m_Helper->sharedPatchSize ();

        // Keep in mind that the patch has its dimentions doubled.
        m_PatchDataSize = m_PatchSize / 2;
        m_nPatchesX = width  / m_PatchDataSize;
        m_nPatchesY = height / m_PatchDataSize;
        if (width  % m_PatchDataSize) m_nPatchesX++;
        if (height % m_PatchDataSize) m_nPatchesY++;

        m_nPatches          = (m_nPatchesX + 2) * (m_nPatchesY + 2);
        m_DataPatches       = new IArray * [m_nPatches];
        m_DataPatches_proxy = new Patch    [m_nPatches];
        if (m_DataPatches &&
            m_DataPatches_proxy) {
            // Generate the patches.
            for (int i = 0; i < m_nPatches; i++) {
                m_DataPatches[i] = g_fac_SPAWN_ARRAY (
                    helper,
                    m_PatchSize,
                    m_PatchSize,
                    format,
                    type,
                    IArray::FLAGS_NULL
                );
/*
                if (m_DataPatches[i])
                    m_DataPatches[i]->zero ();
*/
            }

            // Init the rectangles.
            for (int y = -1, i = 0; y <= m_nPatchesY; y++) {
                for (int x = -1; x <= m_nPatchesX; x++, i++) {
                    m_DataPatches_proxy[i].setData   (m_DataPatches[i]);
                    m_DataPatches_proxy[i].setLeft   (m_PatchDataSize * x);
                    m_DataPatches_proxy[i].setTop    (m_PatchDataSize * y);
                    m_DataPatches_proxy[i].setRight  (m_PatchDataSize * (x + 1));
                    m_DataPatches_proxy[i].setBottom (m_PatchDataSize * (y + 1));
                    m_DataPatches_proxy[i].setCellX  (x);
                    m_DataPatches_proxy[i].setCellY  (y);
                }
            }
        }

        m_ElementSz = m_DataPatches[0]->element_size ();
    } else {
        m_nPatches = 0;
        m_DataPatches       = NULL;
        m_DataPatches_proxy = NULL;
    }
}

ArrayGrid::~ArrayGrid (void) {
    ISpawnedObject::_unregister (this);

    for (int i = 0; i < m_nPatches; i++) {
        SAFE_RELEASE (m_DataPatches[i]);
    }

    SAFE_DELETE_A (m_DataPatches_proxy);
    SAFE_DELETE_A (m_DataPatches);
    m_nPatches = 0;

    SAFE_RELEASE (m_Helper);
}

void ArrayGrid::zero (void) {
    if (m_DataPatches) {
        for (int i = 0; i < m_nPatches; i++) {
#ifdef DEBUG_COLORED_PATCHES
            BYTE r = (BYTE) ((16 * rand ()) / RAND_MAX);
            BYTE g = (BYTE) ((16 * rand ()) / RAND_MAX);
            BYTE b = (BYTE) ((16 * rand ()) / RAND_MAX);

            m_DataPatches[i]->fill (r, g, b, 16);
#else
            m_DataPatches[i]->zero ();
#endif
        }
    }
}

void ArrayGrid::fill (BYTE r, BYTE g, BYTE b, BYTE a) {
    if (m_DataPatches) {
        for (int i = 0; i < m_nPatches; i++) {
            m_DataPatches[i]->fill (r, g, b, a);
        }
    }
}

void ArrayGrid::fill (float r, float g, float b, float a) {
    if (m_DataPatches) {
        for (int i = 0; i < m_nPatches; i++) {
            m_DataPatches[i]->fill (r, g, b, a);
        }
    }
}

IArrayGridPatch * ArrayGrid::patchptr (int x, int y) {
    if (m_DataPatches) {
        if (x >= -1 && x <= m_nPatchesX &&
            y >= -1 && y <= m_nPatchesY) {
            return m_DataPatches_proxy + (x + 1 + (y + 1) * (m_nPatchesX + 2));
        }
    }

    return NULL;
}

int ArrayGrid::touched (int x, int y, int w, int h, IArrayGridPatch ** dest) {
    sRect bliss;
    bliss.left   = x;
    bliss.top    = y;
    bliss.right  = x + w;
    bliss.bottom = y + h;

    return touched (&bliss, dest);
}

int ArrayGrid::touched (const sRect * rect, IArrayGridPatch ** dest) {
    int written = 0;
    if (dest) {
        // Enumerate all patches.
        for (int i = 0; i < m_nPatches; i++) {
            if (QHASPolygon::intersects (rect, m_DataPatches_proxy[i].rect ())) {
                dest[written++] = m_DataPatches_proxy + i;
            }
        }
    }

    return written;
}

IArrayGridPatch * ArrayGrid::touched (int x, int y) {
    for (int i = 0; i < m_nPatches; i++) {
        const sRect * r = m_DataPatches_proxy[i].rect ();
        if (r->left   <= x &&
            r->right  >= x &&
            r->top    <= y &&
            r->bottom >= y) {
            return m_DataPatches_proxy + i;
        }
    }

    return NULL;
}

