/***************************************************************************
 *  Copyright (C) 2009 by Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#pragma once
#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

//#ifdef QT_OPENGL_LIB
#   include "../../../glew/include/GL/glew.h"
#   include "iArray.h"

#   define BUFFER_OFFSET(i) ((char *)NULL + (i))
#   define UINT_UNDEFINED   ((UINT)-1)
//#   define ENABLE_PBO_TRANSFER
//#   define ARRAYS_FILL_WITH_ZEROS


enum eGPU_ARRAY_STATE {
    ARRAY_STATE__EMPTY,
    ARRAY_STATE__HERE,
    ARRAY_STATE__HERE_LOCKED,
    ARRAY_STATE__THERE,
    ARRAY_STATE__THERE_LOCKED
};

/**
 * GPU-located array.
 * Incapsulates the PBO-accelerated CPU<->GPU transfers.
 * Implementation is havily based on tutorials by Dominik Göddeke:
 * http://www.mathematik.uni-dortmund.de/~goeddeke/index.html
 * huge thanks to him.
 */
class CArray_gpu : public IArray,
                   public CRefCounter {
    _GLEW_ENABLED

    /**
     * Current state of the array.
     */
    eGPU_ARRAY_STATE state;

    int    m_Flags;
    int    m_Id;
    BYTE  * locked;         /**< RAM data mirror.       */
    BYTE  * lockedu;        /**< Unaligned pointer.     */
    BYTE  * mapped;         /**< Locked PBO data.       */
    long   block_sz;        /**< Size of array.         */
    long   block_elsz;      /**< Size of array element. */
    long   block_w;         /**< Width of array.        */
    long   block_h;         /**< Height of array.       */
    GLenum dataformat;      /**< Format of texture data.*/
    GLenum transfer_type;   /**<                        */
    GLenum transfer_prec;   /**<                        */
    GLuint tex;             /**< Handle to texture.     */

    /**
     * Helper object.
     */
    IOpenGLHelper * m_Helper;

    /**
     * Set to TRUE if the data,
     * might have been modified
     * during the last lock.
     */
    bool modified;
    bool fbo_is_shared;

    /**
     * Framebuffer object to handle the transfers.
     */
    GLuint fbo;

#ifdef ENABLE_PBO_TRANSFER
    /**
     * Pixel buffer object to speedup our transfers.
     */
    GLuint buf;
    bool   allow_pbo;
#endif

    IFBOWrapper * fbo_w;

    GLenum buf_access;
    GLenum buf_usage;

protected:
    /**
     * Protected destructor.
     */
    virtual ~CArray_gpu (void);

    virtual void suicide (void) {
        delete this;
    }

    /**
     * Import the data fro the
     * given texture.
     */
    void import_tex (GLuint tex);

    void obtain_FBO (void);
    void obtain_PBO (void);

public:
    /**
     * Obviously, it's a constructor.
     */
    CArray_gpu (IOpenGLHelper * helper,
                LONG width, LONG height,
                GLenum format, GLenum type,
                GLuint src,
                int flags);

    /**
     * Lock the array data for software access.
     * If data is currently loaded into some hardware,
     * it will be extracted back. You must call unlock ()
     * to let the data be returned to hardware if needed,
     */
    virtual void * lock_pointer (GLenum access) {
        return lock_pointer (0, 0, 0, 0, access);
    }

    /**
     * Lock the specified data region.
     */
    virtual void * lock_pointer (int x, int y, int w, int h, GLenum access);

    /**
     * Lock the array data for software read-only access.
     * If data is currently loaded into some hardware,
     * it will be extracted back. You must call unlock ()
     * to let the data be returned to hardware if needed,
     */
    virtual const void * lock_pointer_const (void) {
        return lock_pointer_const (0, 0, 0, 0);
    }

    /**
     * Lock the specified data region.
     */
    virtual const void * lock_pointer_const (int x, int y, int w, int h) {
        // FIXME Implement the proper read-only lock.
        const void * p = lock_pointer (x, y, w, h, GL_READ_ONLY_ARB);
        if (p) {
            modified = FALSE;
        }

        return p;
    }

    /**
     * Lock the array data for hardware access.
     * Returns the handle of data loaded into
     * the hardware. If the data is extracted now,
     * it will be rpacked back if not locked.
     * @return Handle of -1 if can't lock or not implemented.
     */
    virtual LONG lock_gltex (void);

    /**
     * Create a new object with the same
     * content.
     */
    virtual IArray * duplicate (void);

    /**
     * Unlock the array data
     * wherever it is located,
     */
    virtual void unlock (void);

    /**
     * The width of array.
     */
    virtual LONG width (void) const {
        return block_w;
    }

    /**
     * The height of array.
     */
    virtual LONG height (void) const {
        return block_h;
    }

    /**
     * Size of a single array element.
     */
    virtual LONG element_size (void) const {
        return block_elsz;
    }

    /**
     * Replace the data in array by
     * providing the handle to new data
     * plane in hardware.
     * @param new_handle Handle to new texture.
     * @return Handle to old texture.
     */
    virtual LONG swap (LONG new_handle);

    /**
     * Fill the array with zeros.
     */
    virtual void zero (void) {
        fill ((BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0);
    }

    /**
     * Fill the array with given data.
     */
    virtual void fill (BYTE r, BYTE g, BYTE b, BYTE a);

    /**
     * Fill the array with given data.
     */
    virtual void fill (float r, float g, float b, float a);

    /**
     * Move all buffer contents to software
     * memory and free all OpenGL-related stuff.
     */
    virtual void purgeToSW (void);

    /**
     * Format of the pixel data returned by lock_pointer ().
     */
    virtual GLenum transferFormat (void) const {
        return transfer_type;
    }

    /**
     * Type of the pixel data returned by lock_pointer ().
     */
    virtual GLenum transferType (void) const {
        return transfer_prec;
    }

    /********* ISPAWNEDOBJECT IMPLANTS *************/

    /**
     * Get the classname for this object.
     */
    virtual const char * myClassName (void) {
        return "Common::CArray_gpu";
    }

    /**
     * Get the Id of current allocated object.
     */
    virtual int objectId (void) {
        return m_Id;
    }

    /**
     * Increase the object's
     * reference counter.
     */
    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    /**
     * Release the object.
     */
    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }
};

//#endif // QT_OPENGL_LIB
