/***
 *  Copyright (C) 2009 Anton R. <commanderkyle@gmail.com>
 *
 *  This file is a part of QAquarelle project.
 *
 *  QAquarelle is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  QAquarelle is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with QAquarelle.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef PIPELINE_STAGE0_H
#define PIPELINE_STAGE0_H

#include "commdefs.h"
#include "commtypes.h"
#include "refcounter.h"

//#ifdef QT_OPENGL_LIB
#   include "../../../glew/include/GL/glew.h"

#include "iLogger.h"
#include "iPipeline.h"
#include <QMutex>


class Pipeline_stage0 : public IPipeline,
                        public CRefCounter {
    _VERBOSEOBJECT
    _GLEW_ENABLED

public:

    /*********** IPIPELINE IMPLANTS ************/

    /**
     * Check if the pipeline accepts the
     * given channel type.
     */
    virtual bool acceptsChannel (FOURCC uid);

    /**
     * Add the given channel to the processing queue.
     * The channel is automatically detached after
     * each iterate () call.
     * /return TRUE if the channel has been accepted.
     */
    virtual bool pushChannel (IChannel * channel);

    /**
     * Pipeline priority controls the position
     * of a given processor in a processing queue.
     */
    virtual int priority (void) {
        return PIPELINE_PRIORITY_LOW;
    }

    /**
     * Run a single iteration on all attached channels.
     * /return TRUE if any channel has been changed.
     */
    virtual bool iterate (float dt);

    /**
     * Reload all OpenGL shaders.
     */
    virtual void reloadShaders (void);

    /************** IPLUGIN IMPLANTS ****************/

    /**
     * Get the name of this plugin.
     */
    virtual const char * name (void) {
        return "PaintStage0";
    }

    /**
     * Get the description of this plugin.
     */
    virtual const char * description (void) {
        return "This pipeline simulates some basic water dynamics.";
    }

    /**
     * Get the author of this plugin.
     */
    virtual const char * author (void) {
        return "shadow-of-kyle";
    }

    /**
     * Get the number of custom parameters
     * accepted by the plugin.
     */
    virtual int customParametersCount (void) const {
        return 7;
    }

    /**
     * Get the name of custom parameter
     * with the given index.
     * /return Pointer to string of NULL if the index is out of range.
     */
    virtual const char * customParameterName (int index) const;

    /**
     * Get the type of custom parameter
     * with the given index.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (int index) const;

    /**
     * Get the type of custom parameter
     * with the given name.
     */
    virtual ePLUGIN_PARAM_TYPE customParameterType (const char * name) const;

    /**
     * Get the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully retrieved.
     */
    virtual bool customParameterGet (const char * name, void * buffer);

    /**
     * Set the value of custom parameter with the given name.
     * /return TRUE if the value has been successfully set.
     */
    virtual bool customParameterSet (const char * name, const void * buffer);

    /**
     * Deinit all OpenGL-related data
     * structures to be prepared for a
     * context switching.
     */
    virtual void deInit (void);

    /********** ISPAWNEDOBJECT IMPLANTS *************/

    virtual int objectId (void) {
        return m_Id;
    }

    virtual void addRef (void) {
        CRefCounter::addRef ();
    }

    virtual void release (void) {
        CRefCounter::release ();
    }

    /**
     * Ask the object to emit all
     * notification signals if any.
     */
    virtual void pokeObject (void) {
    }

    virtual const char * myClassName (void) {
        return "Server::Pipeline_stage0";
    }

    /********** IVERBOSEOBJECT IMPLANTS ***********/

    virtual const char * lastError (void) const {
        return m_LastError;
    }

    virtual void attachLogger (ILogger * logger) {
        SAFE_RELEASE (m_Logger);

        if (logger) {
            m_Logger = logger;
            m_Logger->addRef ();
        }
    }

    virtual void enableLogging (bool enable) {
        m_LoggingEnabled = enable;
    }

    virtual bool loggingEnabled (void) {
        return m_LoggingEnabled;
    }

private:

    int          m_Id;
    bool         m_GLinited;
    bool         m_LastStatus;

    /**
     * Get the parameter index by its name.
     */
    int parameterDehash (const char * name) const;

protected:
    friend class BasicDynamics;

    IOpenGLHelper * m_Helper;
    IChannel      * m_chOptical;
    IChannel      * m_chMeta;
    IArray        * m_Paper;
    float           m_PaperOffset[2];
    float           m_PaperScale[2];

    IOpenGLShader * m_glslCopy;
    IOpenGLShader * m_glslDry0;
    IOpenGLShader * m_glslDry1;
    IOpenGLShader * m_glslAdvect;
    IOpenGLShader * m_glslSelfAdvect;
    GLuint          m_glPar0_src;
    GLuint          m_glPar1_optical;
    GLuint          m_glPar1_dt;
    GLuint          m_glPar1_gran;
    GLuint          m_glPar1_ballance1;
    GLuint          m_glPar1_ballance2;
    GLuint          m_glPar2_source;
    GLuint          m_glPar2_optical_center;
    GLuint          m_glPar2_optical_left;
    GLuint          m_glPar2_optical_top;
    GLuint          m_glPar2_optical_right;
    GLuint          m_glPar2_optical_bottom;
    GLuint          m_glPar2_dt;
    GLuint          m_glPar2_embossing;
    GLuint          m_glPar3_src;
    GLuint          m_glPar3_meta0;
    GLuint          m_glPar3_txx;
    GLuint          m_glPar3_txy;
    GLuint          m_glPar3_dtx;
    GLuint          m_glPar3_dt;
    GLuint          m_glPar4_meta0;
    GLuint          m_glPar4_txx;
    GLuint          m_glPar4_txy;
    GLuint          m_glPar4_dt;

    GLuint          m_FBO;

    float           m_Step;
    float           m_DryingSpeed;
    float           m_FlowSpeed;
    float           m_Granulation;
    float           m_Embossing;

    void detach (void);

    Pipeline_stage0 (IOpenGLHelper * helper);
    virtual ~Pipeline_stage0 (void);

    virtual void suicide (void) {
        delete this;
    }
};

//#endif // QT_OPENGL_LIB
#endif // PIPELINE_STAGE0_H
